Package: GAReg
Type: Package
Title: Genetic Algorithms in Regression
Version: 0.1.2
Authors@R: c(
    person(
    given = "Mo",
    family = "Li",
    role = c("aut", "cre"),
    email = "mo.li@louisiana.edu"),
    person(
    given = "QiQi", 
    family = "Lu", 
    role = "aut", 
    email = "qlu2@vcu.edu"),
    person(
    given = "Robert", 
    family = "Lund", 
    role = "aut", 
    email = "rolund@ucsc.edu"),
    person(
    given = "Xueheng", 
    family = "Shi", 
    role = "aut", 
    email = "shixueheng@gmail.com")
  )
Description: Provides a genetic algorithm framework for regression problems requiring discrete optimization over model spaces with unknown or varying dimension, where gradient-based methods and exhaustive enumeration are impractical. Uses a compact chromosome representation for tasks including spline knot placement and best-subset variable selection, with constraint-preserving crossover and mutation, exact uniform initialization under spacing constraints, steady-state replacement, and optional island-model parallelization from Lu, Lund, and Lee (2010, <doi:10.1214/09-AOAS289>). The computation is built on the 'GA' engine of Scrucca (2017, <doi:10.32614/RJ-2017-008>) and 'changepointGA' engine from Li and Lu (2024, <doi:10.48550/arXiv.2410.15571>). In challenging high-dimensional settings, 'GAReg' enables efficient search and delivers near-optimal solutions when alternative algorithms are not well-justified.
License: Apache License (== 2.0)
RoxygenNote: 7.3.3
Depends: R (>= 4.3.0)
Imports: stats, splines, utils, methods, changepointGA, GA
URL: https://github.com/mli171/GAReg
BugReports: https://github.com/mli171/GAReg/issues
Suggests: MASS, knitr, rmarkdown
Encoding: UTF-8
VignetteBuilder: knitr
LazyData: true
NeedsCompilation: no
Packaged: 2026-04-27 04:48:47 UTC; mli171
Author: Mo Li [aut, cre],
  QiQi Lu [aut],
  Robert Lund [aut],
  Xueheng Shi [aut]
Maintainer: Mo Li <mo.li@louisiana.edu>
Repository: CRAN
Date/Publication: 2026-04-27 05:10:23 UTC
Built: R 4.5.2; ; 2026-04-27 09:35:58 UTC; unix

Package: GAReg
Type: Package
Title: Genetic Algorithms in Regression
Version: 0.1.2
Authors@R: c(
    person(
    given = "Mo",
    family = "Li",
    role = c("aut", "cre"),
    email = "mo.li@louisiana.edu"),
    person(
    given = "QiQi", 
    family = "Lu", 
    role = "aut", 
    email = "qlu2@vcu.edu"),
    person(
    given = "Robert", 
    family = "Lund", 
    role = "aut", 
    email = "rolund@ucsc.edu"),
    person(
    given = "Xueheng", 
    family = "Shi", 
    role = "aut", 
    email = "shixueheng@gmail.com")
  )
Description: Provides a genetic algorithm framework for regression problems requiring discrete optimization over model spaces with unknown or varying dimension, where gradient-based methods and exhaustive enumeration are impractical. Uses a compact chromosome representation for tasks including spline knot placement and best-subset variable selection, with constraint-preserving crossover and mutation, exact uniform initialization under spacing constraints, steady-state replacement, and optional island-model parallelization from Lu, Lund, and Lee (2010, <doi:10.1214/09-AOAS289>). The computation is built on the 'GA' engine of Scrucca (2017, <doi:10.32614/RJ-2017-008>) and 'changepointGA' engine from Li and Lu (2024, <doi:10.48550/arXiv.2410.15571>). In challenging high-dimensional settings, 'GAReg' enables efficient search and delivers near-optimal solutions when alternative algorithms are not well-justified.
License: Apache License (== 2.0)
RoxygenNote: 7.3.3
Depends: R (>= 4.3.0)
Imports: stats, splines, utils, methods, changepointGA, GA
URL: https://github.com/mli171/GAReg
BugReports: https://github.com/mli171/GAReg/issues
Suggests: MASS, knitr, rmarkdown
Encoding: UTF-8
VignetteBuilder: knitr
LazyData: true
NeedsCompilation: no
Packaged: 2026-04-27 04:48:47 UTC; mli171
Author: Mo Li [aut, cre],
  QiQi Lu [aut],
  Robert Lund [aut],
  Xueheng Shi [aut]
Maintainer: Mo Li <mo.li@louisiana.edu>
Repository: CRAN
Date/Publication: 2026-04-27 05:10:23 UTC
Built: R 4.6.0; ; 2026-04-27 05:20:08 UTC; unix

# Generated by roxygen2: do not edit by hand

export(FDRCalc)
export(Popinitial_fixknots)
export(cptgaControl)
export(crossover_fixknots)
export(fixknotsIC)
export(gareg_knots)
export(gareg_subset)
export(mutation_fixknots)
export(selectTau_uniform_exact)
export(splineX)
export(subsetBIC)
export(varyknotsIC)
exportClasses(gareg)
exportMethods(show)
exportMethods(summary)
importClassesFrom(GA,ga)
importClassesFrom(GA,gaisl)
importClassesFrom(changepointGA,cptga)
importClassesFrom(changepointGA,cptgaisl)
importFrom(GA,ga)
importFrom(GA,gaisl)
importFrom(changepointGA,cptga)
importFrom(changepointGA,cptgaisl)
importFrom(methods,new)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,show)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(stats,.lm.fit)
importFrom(stats,glm.control)
importFrom(stats,glm.fit)
importFrom(stats,lm.fit)
importFrom(stats,runif)
importFrom(utils,str)

# GAReg

## Genetic Algorithms in Regression.

Authors: Mo Li (mo.li@louisiana.edu), QiQi Lu (qlu2@vcu.edu), Robert Lund (rolund@ucsc.edu), Xueheng Shi (xshi11@unl.edu)

## Overview

GAReg is a unified genetic algorithm framework for regression 
problems that require discrete optimization over model spaces with unknown or 
variable dimension, where gradient-based methods and exhaustive search are 
impractical. It provides a compact chromosome representation for tasks such 
as optimal spline knot placement, best-subset variable selection, and multiple 
changepoint detection, along with exact uniform initialization, 
constraint-preserving crossover and mutation, steady-state replacement, 
and optional island-model parallelization. In challenging high-dimensional 
settings, GAReg enables efficient search and delivers near-optimal solutions 
when alternative algorithms are not well-justified.

## Package download and installation

You can install the version of changepointGA from CRAN:

```{r}
install.packages("GAReg")
```

or the development version from Github:

```{r}
# install.packages("remotes")
remotes::install_github("mli171/GAReg", build_vignettes = TRUE, dependencies = TRUE)
```

## Open the Vignette in R for more details

```{r}
browseVignettes("GAReg")
```

## GAReg: Genetic Algorithm Applications in Regression

### Best Subset Variable Selection

We use the simulation function below for subset selection illustration. 
Here, `n` is the number of observations and `p` is the number of predictors. 
For the covairates, `s0` represent number of truly active predictors, valued 
range from `0` to `p`. `magnitudes_range` specifies the range of significantly 
expressed coefficients that corresponding to the truly active predictors. 
If `rho` is specified with some values, the autoregressive structure is 
introduced into the error terms. If `rho=NULL`, we will have independent and 
identically distributed (IID) errors. We can also specify `sigma` for the error 
standard deviation.

```{r}
sim_subset_data <- function(n = 60, p = 50, s0 = 25, sigma = 1.5,
                            magnitudes_range = c(0.5, 2),
                            rho = NULL,
                            seed = NULL) {
  stopifnot(n > 0, p > 0, s0 >= 0, s0 <= p, sigma >= 0)
  if (!is.null(seed)) set.seed(seed)

  X <- matrix(rnorm(n * p), n, p)
  
  # Active set and coefficients
  true_idx    <- if (s0 > 0) sort(sample.int(p, s0)) else integer(0)
  signs       <- if (s0 > 0) sample(c(-1, 1), s0, replace = TRUE) else numeric(0)
  magnitudes  <- if (s0 > 0) runif(s0, magnitudes_range[1], magnitudes_range[2]) else numeric(0)

  beta_true <- numeric(p)
  if (s0 > 0) beta_true[true_idx] <- magnitudes * signs

  if (is.null(rho)) {
    e <- rnorm(n, sd = sigma)
  } else {
    sd_innov <- sigma * sqrt(1 - rho^2)
    burn_in  <- 100
    z        <- rnorm(n + burn_in, sd = sd_innov)
    e_full   <- numeric(n + burn_in)
    for (t in 2:(n + burn_in)) e_full[t] <- rho * e_full[t - 1] + z[t]
    e <- e_full[(burn_in + 1):(burn_in + n)]
  }

  y <- as.numeric(X %*% beta_true + e)

  DF <- data.frame(y = y, as.data.frame(X))
  colnames(DF)[-1] <- paste0("X", seq_len(p))

  list(
    X = X,
    y = y,
    beta_true = beta_true,
    true_idx = true_idx,
    DF = DF,
    rho = if (is.null(rho)) NULL else rho,
    args = list(n = n, p = p, s0 = s0, sigma = sigma,
                magnitudes_range = magnitudes_range,
                rho = rho, seed = seed)
  )
}

sim <- sim_subset_data(n=100, p=50, s0=25, sigma=1.5, rho=NULL, seed=123)
y <- sim$y
X <- sim$X
ga <- gareg_subset(y=y, X=X, gaMethod = "GA", monitor = FALSE,
                   gacontrol=list(popSize=120,
                                  maxiter=20000,
                                  run=4000,
                                  pmutation=0.02))
summary(ga)

# False Discovery Rate and True Positive Rate calculation function
res <- FDRCalc(truelabel=sim$true_idx, predlabel=ga@bestsol, N=50)
# FALSE Discover Rate
res$fdr
# TRUE Positive Rate
res$tpr
```

### Changepoint Detection

The multiple changepoint detection can be conducted through `changepointGA` 
package (Li and Lu, 2024). The BIC penalized function is provided below for IID 
data. The related math details can be found in (Li et al., 2026).

```{r}
BIC.cpt = function(chromosome, Xt){
  
  m   = chromosome[1]
  tau = chromosome[2:(2 + m - 1)]
  N   = length(Xt)
  
  if(m==0){
    mu.hat = mean(Xt)
    sigma.hatsq = sum( (Xt-mu.hat)^2 )/N
    BIC.obj = 0.5*N*log(sigma.hatsq)+ 2*log(N)
  }
  else{
    tau.ext = c(1, tau, N+1)
    seg.len = diff(tau.ext)
    ff = rep(0:m, times=seg.len)
    Xseg = split(Xt, ff)
    mu.seg = unlist(lapply(Xseg,mean), use.names=F)
    mu.hat = rep(mu.seg, seg.len)
    sigma.hatsq = sum( (Xt-mu.hat)^2 )/N
    BIC.obj = 0.5*N*log(sigma.hatsq) + (2*m + 2)*log(N)
  }
  return(BIC.obj)
}

# IID data
set.seed(1234)
n = 200
et = rnorm(n)
Xt = et + rep(c(0,2,0,2), each=n/4)

library(changepointGA)
GA.res <- cptga(
  ObjFunc = BIC.cpt, N = n, popSize = 500, 
  pcrossover = 0.95, pmutation = 0.3, pchangepoint = 10/n,
  Xt = Xt
)
summary(GA.res)
```

### Optimal Knot Selection for Splines

The classic motocycle impact dataset from `MASS` package (Venables & Ripley, 2002) 
is used as example here. 

```{r}
library(MASS)
library(splines)

data(mcycle)
head(mcycle)
```

1. Different GA model set-up

- Varying knots with single population GA;

```{r}
g1 <- gareg_knots(
  y=mcycle$accel, x=mcycle$times,
  minDist = 5,
  gaMethod = "cptga",
  cptgactrl = cptgaControl(popSize=200, pcrossover=0.9, pmutation=0.3),
  ic_method = "BIC"
)
summary(g1)

# knots location
g1@bestsol
```

- Varying knots with island model GA;

```{r}
g2 <- gareg_knots(
  y=mcycle$accel, x=mcycle$times,
  minDist = 5,
  gaMethod = "cptgaisl",
  cptgactrl = cptgaControl(numIslands=5, popSize=200, maxMig=250,
                           pcrossover=0.9, pmutation=0.3),
  ic_method = "BIC"
)
summary(g2)
```

- Fixed knots with single population GA;
```{r}
g3 <- gareg_knots(
  y=mcycle$accel, x=mcycle$times,
  fixedknots = 3,
  minDist = 5,
  gaMethod = "cptga",
  cptgactrl = cptgaControl(popSize=200, pcrossover=0.9, pmutation=0.3),
  ic_method = "BIC"
)
summary(g3)
```

- Fixed knots with island model GA;

```{r}
g4 <- gareg_knots(
  y=mcycle$accel, x=mcycle$times,
  fixedknots = 4,
  minDist = 5,
  gaMethod = "cptgaisl",
  cptgactrl = cptgaControl(numIslands=5, popSize=200, maxMig=250, 
                           pcrossover=0.9, pmutation=0.3),
  ic_method = "BIC"
)
summary(g4)
```

- Comparison;

```{r}
y = mcycle$accel
x = mcycle$times
x_unique = unique(x)

tBIC.vary.ga     = g1@bestsol
tBIC.vary.gaisl  = g2@bestsol
tBIC.fix.3.ga    = g3@bestsol
tBIC.fix.4.gaisl = g4@bestsol

bsfit.vary.ga     = lm(y ~ bs(x, knots=x_unique[g1@bestsol], Boundary.knots = range(x)))
bsfit.vary.gaisl  = lm(y ~ bs(x, knots=x_unique[g2@bestsol], Boundary.knots = range(x)))
bsfit.fix.3.ga    = lm(y ~ bs(x, knots=x_unique[g3@bestsol], Boundary.knots = range(x)))
bsfit.fix.4.gaisl = lm(y ~ bs(x, knots=x_unique[g4@bestsol], Boundary.knots = range(x)))

plot(x, y, xlab = "Time (ms)", ylab = "Acceleration (g)")
ht = seq(min(x), max(x), length.out = 200)
lines(ht, predict(bsfit.vary.ga, data.frame(x = ht)), col="blue", lty = 5, lwd = 2)
lines(ht, predict(bsfit.vary.gaisl, data.frame(x = ht)), col="orange", lty = 4, lwd = 2)
lines(ht, predict(bsfit.fix.3.ga, data.frame(x = ht)), col="purple", lty = 3, lwd = 2)
lines(ht, predict(bsfit.fix.4.gaisl, data.frame(x = ht)), col="#D55E00", lty = 2, lwd = 2)
legend("bottomright", 
       legend = c("Varying knots GA", 
                  "Varying knots island model GA", 
                  "Fixed 3 knots GA", 
                  "Fixed 4 knots island model GA"),
       lty = 5:2, lwd = 2,
       col = c("blue", "orange", "purple", "#D55E00"), bty = "n")
```

2. Spline options: piecewise polynomials, natural cubic, and B-splines

This section illustrates how to build spline design matrices via `splineX()` 
for three common options:

- `type="ppolys"`: degree-d truncated power piecewise polynomials;
- `type="ns"`: degree-3 natural cubic spline (degree will be ignored);
- `type="bs"`: degree-d B-spline basis;

We’ll use the motorcycle acceleration data `MASS::mcycle`, create interior 
knots at quantiles of `times`, and compare how different spline types/degrees 
behave. Here, we only illustrate through `Varying number and locations of knots` 
set-up (Let GA choose both **how many** knots and **where** they go). 

```{r}
g_pp3 <- gareg_knots(
  y = y, x = x,
  minDist   = 5,
  gaMethod  = "cptga",
  ObjFunc   = NULL,          # use default varyknotsIC
  type      = "ppolys",
  degree    = 3,             # degree-3 piecewise polynomial
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_pp3)

g_ns <- gareg_knots(
  y = y, x = x,
  minDist   = 5,
  gaMethod  = "cptga",
  type      = "ns",          # natural cubic (degree ignored)
  degree    = 3,             # ignored for "ns"
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_ns)

g_bs1 <- gareg_knots(
  y = y, x = x,
  minDist   = 5,
  gaMethod  = "cptga",
  type      = "bs",
  degree    = 1,             # linear B-splines
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_bs1)
```

# References

Li, M., & Lu, Q. (2024). changepointGA: An R package for Fast Changepoint Detection via Genetic Algorithm. arXiv preprint arXiv:2410.15571.

Mo Li, QiQi Lu, Robert Lund, & Xueheng Shi. (2026). Genetic Algorithms in Regression. arXiv preprint arXiv:.

Venables, W. N. & Ripley, B. D. (2002) Modern Applied Statistics with S. Fourth Edition. Springer, New York. ISBN 0-387-95457-0.


# Code style

Before pushing changes, please run 

```r
styler::style_pkg()
```

to ensure your code follows the tidyverse style guide.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE
)

set.seed(12345)

## ----setup--------------------------------------------------------------------
library(GAReg)

## -----------------------------------------------------------------------------
sim_subset_data <- function(n = 60, p = 50, s0 = 25, sigma = 1.5,
                            magnitudes_range = c(0.5, 2),
                            rho = NULL,
                            seed = NULL) {
  stopifnot(n > 0, p > 0, s0 >= 0, s0 <= p, sigma >= 0)
  if (!is.null(seed)) set.seed(seed)

  X <- matrix(rnorm(n * p), n, p)

  # Active set and coefficients
  true_idx <- if (s0 > 0) sort(sample.int(p, s0)) else integer(0)
  signs <- if (s0 > 0) sample(c(-1, 1), s0, replace = TRUE) else numeric(0)
  magnitudes <- if (s0 > 0) runif(s0, magnitudes_range[1], magnitudes_range[2]) else numeric(0)

  beta_true <- numeric(p)
  if (s0 > 0) beta_true[true_idx] <- magnitudes * signs

  if (is.null(rho)) {
    e <- rnorm(n, sd = sigma)
  } else {
    sd_innov <- sigma * sqrt(1 - rho^2)
    burn_in <- 100
    z <- rnorm(n + burn_in, sd = sd_innov)
    e_full <- numeric(n + burn_in)
    for (t in 2:(n + burn_in)) e_full[t] <- rho * e_full[t - 1] + z[t]
    e <- e_full[(burn_in + 1):(burn_in + n)]
  }

  y <- as.numeric(X %*% beta_true + e)

  DF <- data.frame(y = y, as.data.frame(X))
  colnames(DF)[-1] <- paste0("X", seq_len(p))

  list(
    X = X,
    y = y,
    beta_true = beta_true,
    true_idx = true_idx,
    DF = DF,
    rho = if (is.null(rho)) NULL else rho,
    args = list(
      n = n, p = p, s0 = s0, sigma = sigma,
      magnitudes_range = magnitudes_range,
      rho = rho, seed = seed
    )
  )
}

## -----------------------------------------------------------------------------
sim <- sim_subset_data(n = 100, p = 50, s0 = 25, sigma = 1.5, rho = NULL, seed = 123)
y <- sim$y
X <- sim$X
ga <- gareg_subset(
  y = y, X = X, gaMethod = "GA", monitor = FALSE,
  gacontrol = list(
    popSize = 120,
    maxiter = 20000,
    run = 4000,
    pmutation = 0.02
  )
)
summary(ga)

res <- FDRCalc(truelabel = sim$true_idx, predlabel = ga@bestsol, N = 50)
# FALSE Discover Rate
res$fdr
# TRUE Positive Rate
res$tpr

## -----------------------------------------------------------------------------
BIC.cpt <- function(chromosome, Xt) {
  m <- chromosome[1]
  tau <- chromosome[2:(2 + m - 1)]
  N <- length(Xt)

  if (m == 0) {
    mu.hat <- mean(Xt)
    sigma.hatsq <- sum((Xt - mu.hat)^2) / N
    BIC.obj <- 0.5 * N * log(sigma.hatsq) + 2 * log(N)
  } else {
    tau.ext <- c(1, tau, N + 1)
    seg.len <- diff(tau.ext)
    ff <- rep(0:m, times = seg.len)
    Xseg <- split(Xt, ff)
    mu.seg <- unlist(lapply(Xseg, mean), use.names = F)
    mu.hat <- rep(mu.seg, seg.len)
    sigma.hatsq <- sum((Xt - mu.hat)^2) / N
    BIC.obj <- 0.5 * N * log(sigma.hatsq) + (2 * m + 2) * log(N)
  }
  return(BIC.obj)
}

# IID data
set.seed(1234)
n <- 200
et <- rnorm(n)
Xt <- et + rep(c(0, 2, 0, 2), each = n / 4)

library(changepointGA)
GA.res <- cptga(
  ObjFunc = BIC.cpt, N = n, popSize = 500,
  pcrossover = 0.95, pmutation = 0.3, pchangepoint = 10 / n,
  Xt = Xt
)
summary(GA.res)

## -----------------------------------------------------------------------------
library(MASS)
library(splines)

data(mcycle)
head(mcycle)

## -----------------------------------------------------------------------------
g1 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  minDist = 5,
  gaMethod = "cptga",
  cptgactrl = cptgaControl(popSize = 200, pcrossover = 0.9, pmutation = 0.3),
  ic_method = "BIC"
)
summary(g1)

# knots location
g1@bestsol

## -----------------------------------------------------------------------------
g2 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  minDist = 5,
  gaMethod = "cptgaisl",
  cptgactrl = cptgaControl(
    numIslands = 5, popSize = 200, maxMig = 250,
    pcrossover = 0.9, pmutation = 0.3
  ),
  ic_method = "BIC"
)
summary(g2)

## -----------------------------------------------------------------------------
g3 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  fixedknots = 3,
  minDist = 5,
  gaMethod = "cptga",
  cptgactrl = cptgaControl(popSize = 200, pcrossover = 0.9, pmutation = 0.3),
  ic_method = "BIC"
)
summary(g3)

## -----------------------------------------------------------------------------
g4 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  fixedknots = 4,
  minDist = 5,
  gaMethod = "cptgaisl",
  cptgactrl = cptgaControl(
    numIslands = 5, popSize = 200, maxMig = 250,
    pcrossover = 0.9, pmutation = 0.3
  ),
  ic_method = "BIC"
)
summary(g4)

## ----fig.width=7, fig.height=5, out.width="90%", fig.align='center'-----------
y <- mcycle$accel
x <- mcycle$times
x_unique <- unique(x)

tBIC.vary.ga <- g1@bestsol
tBIC.vary.gaisl <- g2@bestsol
tBIC.fix.3.ga <- g3@bestsol
tBIC.fix.4.gaisl <- g4@bestsol

bsfit.vary.ga <- lm(y ~ bs(x, knots = x_unique[g1@bestsol], Boundary.knots = range(x)))
bsfit.vary.gaisl <- lm(y ~ bs(x, knots = x_unique[g2@bestsol], Boundary.knots = range(x)))
bsfit.fix.3.ga <- lm(y ~ bs(x, knots = x_unique[g3@bestsol], Boundary.knots = range(x)))
bsfit.fix.4.gaisl <- lm(y ~ bs(x, knots = x_unique[g4@bestsol], Boundary.knots = range(x)))

plot(x, y, xlab = "Time (ms)", ylab = "Acceleration (g)")
ht <- seq(min(x), max(x), length.out = 200)
lines(ht, predict(bsfit.vary.ga, data.frame(x = ht)), col = "blue", lty = 5, lwd = 2)
lines(ht, predict(bsfit.vary.gaisl, data.frame(x = ht)), col = "orange", lty = 4, lwd = 2)
lines(ht, predict(bsfit.fix.3.ga, data.frame(x = ht)), col = "purple", lty = 3, lwd = 2)
lines(ht, predict(bsfit.fix.4.gaisl, data.frame(x = ht)), col = "#D55E00", lty = 2, lwd = 2)
legend("bottomright",
  legend = c(
    "Varying knots GA",
    "Varying knots island model GA",
    "Fixed 3 knots GA",
    "Fixed 4 knots island model GA"
  ),
  lty = 5:2, lwd = 2,
  col = c("blue", "orange", "purple", "#D55E00"), bty = "n"
)

## ----vary-ppolys--------------------------------------------------------------
g_pp3 <- gareg_knots(
  y = y, x = x,
  minDist = 5,
  gaMethod = "cptga",
  ObjFunc = NULL, # use default varyknotsIC
  type = "ppolys",
  degree = 3, # degree-3 piecewise polynomial
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_pp3)

## ----vary-ns------------------------------------------------------------------
g_ns <- gareg_knots(
  y = y, x = x,
  minDist = 5,
  gaMethod = "cptga",
  type = "ns", # natural cubic (degree ignored)
  degree = 3, # ignored for "ns"
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_ns)

## ----vary-bs------------------------------------------------------------------
g_bs1 <- gareg_knots(
  y = y, x = x,
  minDist = 5,
  gaMethod = "cptga",
  type = "bs",
  degree = 1, # linear B-splines
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_bs1)

## -----------------------------------------------------------------------------
sessionInfo()

#' Northern Hemisphere temperature reconstruction
#'
#' The 2,000-Year Northern Hemisphere Temperature Reconstruction dataset
#' (Moberg et al., 2005) provides annual temperature anomalies for the
#' Northern Hemisphere from AD 1 to 1979, relative to the 1961–1990 mean.
#' The reconstruction combines high-resolution proxy data (e.g., tree rings)
#' with low-resolution proxies (e.g., sediments) using a wavelet-based method
#' to capture variability across multiple time scales.
#'
#' @format A data frame with 2 variables:
#' \describe{
#'   \item{Year}{Year AD}
#'   \item{Temp}{Temperature anomaly relative to the 1961--1990 mean}
#' }
#' @source Moberg A, Sonechkin DM, Holmgren K, Datsenko NM, Karlén W.
#'   2,000-Year Northern Hemisphere Temperature Reconstruction. IGBP
#'   PAGES/World Data Center for Paleoclimatology Data Contribution Series
#'   #2005-019. NOAA/NGDC Paleoclimatology Program, Boulder, Colorado, USA.
#'
#' @references Moberg A, Sonechkin DM, Holmgren K, Datsenko NM, Karlén W.
#'   (2005). Highly variable Northern Hemisphere temperatures reconstructed
#'   from low- and high-resolution proxy data. *Nature*, 433(7026), 613--617.
"nhtr2005"

#' @importFrom changepointGA cptga cptgaisl
#' @importClassesFrom changepointGA cptga cptgaisl
#' @importFrom GA ga gaisl
#' @importClassesFrom GA ga gaisl
#' @importFrom methods show setGeneric setMethod slot slotNames new
#' @importFrom stats runif lm.fit
#' @importFrom utils str
NULL

#' Genetic-Algorithm–based Optimal Knot Selection
#'
#' @description
#' Runs a GA-based search for changepoints/knots and returns a compact
#' \code{"gareg"} S4 result that stores the backend GA fit
#' (\code{"cptga"} or \code{"cptgaisl"}) plus the essential run settings.
#'
#' @param y Numeric vector of responses (length \code{N}).
#' @param x Optional index/time vector aligned with \code{y}. If missing, it
#'   defaults to \code{seq_along(y)}. Used to derive \code{x_unique} (candidate
#'   knot positions) and passed to the objective function; the GA backend itself
#'   does not use \code{x} directly.
#' @param ObjFunc Objective function or its name. If \code{NULL}, a default
#'   is chosen:
#'   \itemize{
#'     \item \code{fixknotsIC} when \code{fixedknots} is supplied;
#'     \item \code{varyknotsIC} otherwise.
#'   }
#'   A custom function must accept the chromosome and needed data via
#'   named arguments (see the defaults for a template function).
#' @param fixedknots \code{NULL} (varying-knots search) or an integer giving the
#'   number of interior knots for a fixed-\eqn{m} search. If non-\code{NULL}, the
#'   method is \code{"fixknots"} and specialized operators are injected unless
#'   overridden in \code{cptgactrl}.
#' @param minDist Integer minimum distance between adjacent changepoints.
#'   If omitted (\code{missing()} or \code{NULL}), the value in \code{cptgactrl}
#'   is used. If supplied here, it overrides the control value.
#' @param degree Integer polynomial degree for \code{"ppolys"} and \code{"bs"}.
#'   Ignored for \code{"ns"} (always cubic). Must be provided for \code{"ppolys"}
#'   and \code{"bs"}.
#' @param type One of \code{c("ppolys", "ns", "bs")}:
#'   piecewise polynomials, natural cubic, or B-spline. See \code{\link{splineX}}.
#'   The first option of `ppolys` is taken by default.
#' @param intercept Logical; include intercept column where applicable.
#'   Default: \code{TRUE}.
#' @param gaMethod GA backend to call: function or name. Supports
#'   \code{"cptga"} (single population) and \code{"cptgaisl"} (islands).
#' @param cptgactrl Control list built with \code{\link{cptgaControl}()}
#'   (or a named list of overrides). When \code{gaMethod = "cptgaisl"},
#'   island-specific knobs like \code{numIslands} and \code{maxMig} are
#'   recognized. Other genetic algorithm parameters can be found in \link[changepointGA:cptga]{cptga} and
#'         \link[changepointGA:cptgaisl]{cptgaisl}.
#' @param monitoring Logical; print short progress messages (also forwarded
#'   into the backend control).
#' @param seed Optional RNG seed; also stored into the backend control.
#' @param ... Additional arguments passed to the GA backend. If the backend
#'   does not accept \code{...}, unknown arguments are silently dropped
#'   (the call is filtered against the backend formals).
#'
#' @details
#' \strong{Engine selection and controls.}
#' The function detects the engine from \code{gaMethod} and constructs a
#' matching control via \code{\link{cptgaControl}()}:
#' \itemize{
#'   \item \code{"cptga"} uses \code{.cptga.default}.
#'   \item \code{"cptgaisl"} uses \code{.cptgaisl.default} (supports
#'         \code{numIslands}, \code{maxMig}, etc.).
#'   \item see other details in \link[changepointGA:cptga]{cptga} and
#'         \link[changepointGA:cptgaisl]{cptgaisl}.
#' }
#' Top-level \code{monitoring}, \code{seed}, and \code{minDist} given to
#' \code{gareg_knots()} take precedence over the control list.
#'
#' \strong{Fix-knots operators.}
#' When \code{fixedknots} is provided and the control does not already
#' override them, the following operators are injected:
#' \code{Popinitial_fixknots}, \code{crossover_fixknots}, \code{mutation_fixknots}.
#'
#' \strong{Spline basis options.}
#' To build spline design matrices (via \code{\link{splineX}}):
#' \itemize{
#'   \item \code{type = "ppolys"}: Degree-\eqn{d} regression spline via truncated-power piecewise polynomials.
#'   \item \code{type = "ns"}: Degree-3 natural cubic spline with zero second-derivative at boundaries.
#'   \item \code{type = "bs"}: Degree-\eqn{d} B-spline basis (unconstrained).
#' }
#'
#' @return An object of class \code{"gareg"} with key slots:
#' \itemize{
#'   \item \code{call}, \code{method} (\code{"varyknots"} or \code{"fixknots"}), \code{N}.
#'   \item \code{objFunc}, \code{gaMethod}, \code{gaFit} (class \code{"cptga"} or \code{"cptgaisl"}), \code{ctrl}.
#'   \item \code{fixedknots}, \code{minDist}, \code{polydegree}, \code{type}, \code{intercept}.
#'   \item \code{bestFitness}, \code{bestChrom}, \code{bestnumbsol}, \code{bestsol}.
#' }
#' Use \code{summary(g)} to print GA settings and the best solution (extracted
#' from \code{g@gaFit}); \code{show(g)} prints a compact header.
#'
#' @section Argument precedence:
#' Values are combined as \emph{control < core < \code{...}}. That is,
#' \code{cptgactrl} provides defaults, then core arguments from
#' \code{gareg_knots()} override those, and finally any matching names in
#' \code{...} override both.
#'
#' @seealso \code{\link{cptgaControl}}, \code{changepointGA::cptga},
#'   \code{changepointGA::cptgaisl}, \code{\link{fixknotsIC}}, \code{\link{varyknotsIC}}
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' N <- 120
#' y <- c(rnorm(40, 0), rnorm(40, 3), rnorm(40, 0))
#' x <- seq_len(N)
#'
#' # 1) Varying-knots with single-pop GA
#' g1 <- gareg_knots(
#'   y, x,
#'   minDist = 5,
#'   gaMethod = "cptga",
#'   cptgactrl = cptgaControl(popSize = 150, pcrossover = 0.9, maxgen = 500)
#' )
#' summary(g1)
#'
#' # 2) Fixed knots (operators auto-injected unless overridden)
#' g2 <- gareg_knots(
#'   y, x,
#'   fixedknots = 5,
#'   minDist = 5
#' )
#' summary(g2)
#'
#' # 3) Island GA with island-specific controls
#' g3 <- gareg_knots(
#'   y, x,
#'   gaMethod = "cptgaisl",
#'   minDist = 6,
#'   cptgactrl = cptgaControl(
#'     engine = "cptgaisl",
#'     numIslands = 8, maxMig = 250,
#'     popSize = 120, pcrossover = 0.9
#'   )
#' )
#' summary(g3)
#' }
#'
#' @export
gareg_knots <- function(y,
                        x,
                        ObjFunc = NULL,
                        fixedknots = NULL,
                        minDist = 3L,
                        degree = 3L,
                        type = c("ppolys", "ns", "bs"),
                        intercept = TRUE,
                        gaMethod = "cptga",
                        cptgactrl = NULL,
                        monitoring = FALSE,
                        seed = NULL,
                        ...) {
  call <- match.call()
  type <- match.arg(type)
  dots <- list(...)

  if (missing(x) || is.null(x)) x <- seq_along(y)
  x_unique <- sort(unique(x))
  n <- length(x_unique)

  gareg_method <- if (is.null(fixedknots)) "varyknots" else "fixknots"
  ga_name <- if (is.function(gaMethod)) deparse(substitute(gaMethod)) else as.character(gaMethod)
  ga_fun <- if (is.function(gaMethod)) gaMethod else get(gaMethod, mode = "function")
  engine <- if (tolower(ga_name) == "cptgaisl") "cptgaisl" else "cptga"

  cptgactrl <- if (is.null(cptgactrl)) {
    cptgaControl(engine = engine)
  } else if (inherits(cptgactrl, "cptgaControl")) {
    cptgaControl(.list = unclass(cptgactrl), engine = engine)
  } else if (is.list(cptgactrl)) {
    cptgaControl(.list = cptgactrl, engine = engine)
  } else {
    stop("'cptgactrl' must be cptgaControl() or a named list.")
  }

  if (!missing(monitoring) && !is.null(monitoring)) cptgactrl$monitoring <- monitoring
  if (!missing(seed) && !is.null(seed)) cptgactrl$seed <- seed

  if (missing(minDist) || is.null(minDist)) {
    if (!is.null(cptgactrl$minDist)) minDist <- cptgactrl$minDist
  } else {
    cptgactrl$minDist <- NULL
  }
  if (!is.null(cptgactrl$seed)) set.seed(cptgactrl$seed)

  if (is.null(ObjFunc)) {
    if (!is.null(fixedknots)) {
      if (monitoring) cat("\nDefault Objective Function (fixknotsBIC) in use ...")
      ObjFunc <- fixknotsIC
    } else {
      if (monitoring) cat("\nDefault Objective Function (varyknotsBIC) in use ...")
      ObjFunc <- varyknotsIC
    }
  } else {
    if (monitoring) cat("\nSelf-defined Objective Function in use ...")
    if (is.character(ObjFunc)) ObjFunc <- get(ObjFunc, mode = "function")
    if (!is.function(ObjFunc)) stop("ObjFunc must be a function or a function name.")
  }

  if (!is.null(fixedknots)) {
    if (is.null(cptgactrl$popInitialize) || identical(cptgactrl$popInitialize, .cptga.default$popInitialize)) {
      cptgactrl$popInitialize <- "Popinitial_fixknots"
    }
    if (is.null(cptgactrl$crossover) || identical(cptgactrl$crossover, .cptga.default$crossover)) {
      cptgactrl$crossover <- "crossover_fixknots"
    }
    if (is.null(cptgactrl$mutation) || identical(cptgactrl$mutation, .cptga.default$mutation)) {
      cptgactrl$mutation <- "mutation_fixknots"
    }
  }

  core_args <- list(
    ObjFunc     = ObjFunc,
    N           = n,
    minDist     = minDist,
    y           = y,
    x           = x,
    x_unique    = x_unique,
    degree      = degree,
    type        = type,
    intercept   = intercept
  )
  if (!is.null(fixedknots)) core_args$fixedknots <- fixedknots

  ga_args <- utils::modifyList(cptgactrl, core_args, keep.null = TRUE)
  if (length(dots)) ga_args <- utils::modifyList(ga_args, dots, keep.null = TRUE)

  ga_formals <- try(names(formals(ga_fun)), silent = TRUE)
  if (inherits(ga_formals, "try-error") || is.null(ga_formals)) ga_formals <- character(0)
  obj_formals <- try(names(formals(ObjFunc)), silent = TRUE)
  if (inherits(obj_formals, "try-error") || is.null(obj_formals)) obj_formals <- character(0)

  if ("..." %in% ga_formals) {
    keep_names <- union(ga_formals, "ObjFunc")
    keep_names <- union(keep_names, intersect(names(ga_args), obj_formals))
    ga_args <- ga_args[intersect(names(ga_args), keep_names)]
  } else {
    ga_args <- ga_args[intersect(names(ga_args), ga_formals)]
  }

  GA.res <- do.call(ga_fun, ga_args)

  object <- new("gareg",
    call        = call,
    method      = gareg_method,
    N           = n,
    objFunc     = ObjFunc,
    gaMethod    = ga_name,
    gaFit       = GA.res,
    fixedknots  = fixedknots,
    minDist     = minDist,
    polydegree  = degree,
    type        = type,
    intercept   = intercept,
    bestFitness = GA.res@overbestfit,
    bestChrom   = GA.res@overbestchrom
  )

  mhat <- object@bestnumbsol <- object@bestChrom[1]
  if (mhat == 0) {
    object@bestsol <- NULL
  } else {
    object@bestsol <- object@bestChrom[2:(1 + mhat)]
  }

  return(object)
}


#' Information criterion for spline regression with a variable number of knots
#'
#' Evaluates an information criterion (BIC, AIC, or AICc) for a regression of
#' \code{y} on a spline basis of \code{x} where the number and locations of
#' interior knots are encoded in the chromosome. Designed for use as a GA
#' objective/fitness function. The spline basis is constructed via [splineX()].
#'
#' @param knot_bin Integer vector (chromosome). Gene 1 stores m, the number of
#'   interior knots. Genes 2:(1+m) are indices into \code{x_unique} selecting the m
#'   interior knots, followed by a sentinel equal to \code{length(x_unique)+1}.
#'   Only genes strictly before the first occurrence of \code{length(x_unique)+1}
#'   are treated as interior indices; genes after the sentinel are ignored. Interior
#'   indices must be in \code{2:(length(x_unique)-1)}, finite, and non-duplicated.
#' @param plen Unused placeholder kept for API compatibility; ignored.
#' @param y Numeric response vector of length \eqn{n}.
#' @param x Numeric predictor (same length as \code{y}) on which the spline
#'   basis is constructed.
#' @param x_unique Optional numeric vector of unique candidate knot locations.
#'   If missing or \code{NULL}, defaults to \code{sort(unique(x))}. Must have
#'   at least three values (two boundaries + one interior) to allow any knots.
#' @param x_base Optional matrix (or vector) of additional covariates to include
#'   linearly alongside the spline basis; coerced to a matrix if supplied.
#' @param degree Integer polynomial degree for \code{type="ppolys"} and
#'   \code{type="bs"} (default \code{3L}). Ignored for \code{type="ns"} (always cubic).
#' @param type One of \code{c("ppolys","ns","bs")}; forwarded to [splineX()].
#' @param intercept Logical; forwarded to [splineX()]. For \eqn{m>0}, the spline
#'   block is \code{splineX(..., intercept=intercept)} and \emph{no explicit 1-column}
#'   is added here; for \eqn{m=0}, an explicit intercept is added via
#'   \code{cbind(1, x_base)}. Set \code{intercept=FALSE} if you plan to add your own 1-column.
#' @param ic_method Which information criterion to return: \code{"BIC"},
#'   \code{"AIC"}, or \code{"AICc"}.
#'
#' @details
#' If \eqn{m = 0}, the model is a pure-linear baseline using only an intercept
#' and \code{x_base}: \code{X <- cbind(1, x_base)} (no spline terms).
#' For \eqn{m > 0}, the spline block is built with [splineX()] using the selected
#' interior knots, with \code{X <- cbind(splineX(..., intercept=intercept), x_base)}.
#'
#' The criteria are computed as:
#' \deqn{\mathrm{BIC} = n \log(\mathrm{SSRes}/n) + p \log n,}
#' \deqn{\mathrm{AIC} = n \log(\mathrm{SSRes}/n) + 2p,}
#' \deqn{\mathrm{AICc} = n \log(\mathrm{SSRes}/n) + 2p +
#'       \frac{2p(p+1)}{n-p-1},}
#' where \eqn{\mathrm{SSRes}} is the residual sum of squares and \eqn{p} is the
#' number of columns in the design matrix \code{X}.
#'
#' @return A single numeric value: the requested information criterion (lower is
#'   better). Returns \code{Inf} for invalid chromosomes/inputs.
#'
#' @note This function allows \eqn{m=0} (no spline terms) so that the GA can
#'   compare against a pure-linear baseline (intercept + \code{x_base}).
#'   Spacing constraints (e.g., minimum distance between indices) should be
#'   enforced by the GA operators or an external penalty.
#'
#' @seealso [fixknotsIC()], [splineX()], \code{\link[splines]{bs}}, \code{\link[splines]{ns}}
#'
#' @examples
#' ## Example with 'mcycle' data (MASS)
#' # y <- mcycle$accel; x <- mcycle$times
#' # x_unique <- sort(unique(x))
#' # chrom <- c(5, 24, 30, 46, 49, 69, length(x_unique) + 1)
#' # varyknotsIC(chrom, y=y, x=x, x_unique=x_unique,
#' #             type="ppolys", degree=3, ic_method="BIC")
#' @export
varyknotsIC <- function(knot_bin,
                        plen = 0,
                        y,
                        x,
                        x_unique,
                        x_base = NULL,
                        degree = 3L,
                        type = c("ppolys", "ns", "bs"),
                        intercept = TRUE,
                        ic_method = "BIC") {
  ic_method <- match.arg(ic_method, c("BIC", "AIC", "AICc"))
  type <- match.arg(type)

  stopifnot(is.numeric(y), is.numeric(x), length(y) == length(x))
  if (!is.null(x_base)) x_base <- as.matrix(x_base)

  if (missing(x_unique) || is.null(x_unique)) {
    x_unique <- sort(unique(x))
  } else {
    x_unique <- sort(unique(x_unique))
  }
  if (length(x_unique) < 3L) {
    return(Inf)
  } # need at least one interior candidate

  n <- length(y)
  m <- as.integer(knot_bin[1L])
  xr <- range(x, na.rm = TRUE)
  Nu <- length(x_unique)

  if (m == 0L) {
    # no knots: intercept + x_base
    tail <- as.integer(knot_bin[-1L])
    if (!any(tail == Nu + 1L)) {
      return(Inf)
    }
    X <- cbind(rep(1, n), x_base)
  } else {
    tail <- as.integer(knot_bin[-1L])
    end_pos <- match(Nu + 1L, tail)
    if (is.na(end_pos)) {
      return(Inf)
    }
    idx <- tail[seq_len(end_pos - 1L)]
    idx <- idx[idx != 0L]

    if (length(idx) != m) {
      return(Inf)
    }

    L <- 2L
    U <- length(x_unique) - 1L
    if (U < L) {
      return(Inf)
    }
    if (any(!is.finite(idx)) || any(idx < L) || any(idx > U)) {
      return(Inf)
    }
    if (anyDuplicated(idx)) {
      return(Inf)
    }

    knotvec <- x_unique[idx]
    if (anyNA(knotvec)) {
      return(Inf)
    }
    if (any(!(knotvec > xr[1] & knotvec < xr[2]))) {
      return(Inf)
    }

    x_spline <- splineX(
      x,
      knots = sort(knotvec), degree = degree, type = type, intercept = intercept
    )
    X <- cbind(x_spline, x_base)
  }

  fit <- stats::.lm.fit(X, y)
  SSRes <- sum(fit$residuals^2)
  p <- NCOL(X)

  switch(ic_method,
    BIC = n * log(SSRes / n) + p * log(n),
    AIC = n * log(SSRes / n) + 2 * p,
    AICc = if (n - p - 1 > 0) {
      n * log(SSRes / n) + 2 * p + (2 * p * (p + 1)) / (n - p - 1)
    } else {
      Inf
    }
  )
}

#' Information criterion for a fixed–knot spline regression
#'
#' Computes an information criterion (BIC, AIC, or AICc) for a regression of
#' \code{y} on a spline basis of \code{x} when the number of interior knots is
#' fixed. This is designed to be used as a fitness/objective function inside a
#' GA search where the chromosome encodes the indices of the interior knots.
#'
#' @param knot_bin Integer vector (chromosome). Gene 1 stores m, the number of
#'   interior knots. Genes 2:(1+m) are indices into \code{x_unique} selecting the m
#'   interior knots, followed by a sentinel equal to \code{length(x_unique)+1}.
#'   Only genes strictly before the first occurrence of \code{length(x_unique)+1}
#'   are treated as interior indices; genes after the sentinel are ignored. Interior
#'   indices must be in \code{2:(length(x_unique)-1)}, finite, and non-duplicated.
#' @param plen Unused placeholder kept for API compatibility with other
#'   objective functions. Ignored.
#' @param y Numeric response vector of length \eqn{n}.
#' @param x Numeric predictor (same length as \code{y}) on which the spline
#'   basis is built.
#' @param x_unique Optional numeric vector of unique candidate knot locations.
#'   If \code{NULL} or missing, it defaults to \code{sort(unique(x))}. Must
#'   contain at least \eqn{m + 2} values (interior + two boundaries).
#' @param x_base Optional matrix (or vector) of additional covariates to include
#'   linearly alongside the spline basis. If supplied, it is coerced to a matrix
#'   and column-bound to the design.
#' @param fixedknots Integer \eqn{m}: the number of interior knots to use.
#'   Internally this determines how many indices are read from \code{knot_bin}.
#' @param degree Integer polynomial degree for \code{type="ppolys"} and
#'   \code{type="bs"} (default \code{3L}). Ignored for \code{type="ns"} (cubic).
#' @param type One of \code{c("ppolys","ns","bs")}; forwarded to [splineX()].
#' @param intercept Logical; forwarded to [splineX()]. For \eqn{m>0}, the spline
#'   block is \code{splineX(..., intercept=intercept)} and no explicit 1-column is
#'   added here. If you add your own intercept in \code{X}, call
#'   \code{splineX(..., intercept=FALSE)}.
#' @param ic_method Character; which information criterion to return:
#'   \code{"BIC"}, \code{"AIC"}, or \code{"AICc"}.
#'
#' @details
#' We decode the interior indices up to the sentinel \code{length(x_unique)+1},
#' validate them (finite, interior, non-duplicated), sort the resulting knot
#' locations internally, and build the design as
#' \code{X <- cbind(splineX(..., intercept=intercept), x_base)}.
#' Invalid chromosomes/inputs return \code{Inf}.
#'
#' @return A single numeric value: the requested information criterion. Lower is
#'   better. Returns \code{Inf} for invalid chromosomes/inputs.
#'
#' @seealso [varyknotsIC()], [splineX()], \code{\link[splines]{bs}}, \code{\link[splines]{ns}}
#'
#' @examples
#' library(MASS)
#' y <- mcycle$accel
#' x <- mcycle$times
#' x_unique <- sort(unique(x))
#' # chromosome encoding 5 interior knot indices with sentinel:
#' chrom <- c(5, 24, 30, 46, 49, 69, length(x_unique) + 1)
#' fixknotsIC(chrom,
#'   y = y, x = x, x_unique = x_unique,
#'   fixedknots = 5, ic_method = "BIC"
#' )
#' @export
fixknotsIC <- function(knot_bin,
                       plen = 0,
                       y,
                       x,
                       x_unique,
                       x_base = NULL,
                       fixedknots,
                       degree = 3L,
                       type = c("ppolys", "ns", "bs"),
                       intercept = TRUE,
                       ic_method = "BIC") {
  ic_method <- match.arg(ic_method)
  type <- match.arg(type)

  stopifnot(is.numeric(y), is.numeric(x), length(y) == length(x))
  if (!is.null(x_base)) x_base <- as.matrix(x_base)

  if (missing(x_unique) || is.null(x_unique)) {
    x_unique <- sort(unique(x))
  } else {
    x_unique <- sort(unique(x_unique))
  }
  # need m interior points and 2 boundaries
  if (length(x_unique) < as.integer(fixedknots) + 2L) {
    return(Inf)
  }

  n <- length(y)
  m <- as.integer(fixedknots)
  xr <- range(x, na.rm = TRUE)
  Nu <- length(x_unique)

  tail <- as.integer(knot_bin[-1L])
  end_pos <- match(Nu + 1L, tail)
  if (is.na(end_pos)) {
    return(Inf)
  }

  cand <- tail[seq_len(end_pos - 1L)]
  cand <- cand[cand != 0L]

  if (length(cand) < m) {
    return(Inf)
  }
  idx <- cand[seq_len(m)]

  L <- 2L
  U <- length(x_unique) - 1L
  if (U < L) {
    return(Inf)
  }
  if (any(!is.finite(idx)) || any(idx < L) || any(idx > U)) {
    return(Inf)
  }
  if (anyDuplicated(idx)) {
    return(Inf)
  }

  knotvec <- x_unique[idx]
  if (anyNA(knotvec)) {
    return(Inf)
  }
  if (any(!(knotvec > xr[1] & knotvec < xr[2]))) {
    return(Inf)
  }

  x_spline <- splineX(
    x,
    knots = sort(knotvec), degree = degree, type = type, intercept = intercept
  )
  X <- cbind(x_spline, x_base)

  fit <- stats::.lm.fit(X, y)
  SSRes <- sum(fit$residuals^2)
  p <- NCOL(X)

  val <- switch(ic_method,
    BIC  = n * log(SSRes / n) + p * log(n),
    AIC  = n * log(SSRes / n) + p * 2,
    AICc = if (n - p - 1 > 0) n * log(SSRes / n) + p * 2 + (2 * p * (p + 1)) / (n - p - 1) else Inf
  )

  return(val)
}


.is_scalar <- function(x) is.atomic(x) && length(x) == 1L
.is_intish <- function(x) is.numeric(x) && length(x) == 1L && is.finite(x) && abs(x - round(x)) < 1e-9
.chk_prob <- function(x, nm) {
  if (!(is.numeric(x) && length(x) == 1L && is.finite(x) && x >= 0 && x <= 1)) {
    stop(sprintf("`%s` must be a single number in [0,1].", nm), call. = FALSE)
  }
}

.validate_ctrl <- function(x, engine = c("cptga", "cptgaisl")) {
  engine <- match.arg(engine)
  if (!.is_intish(x$popSize) || x$popSize < 1) stop("`popSize` must be integer >= 1.")
  .chk_prob(x$pcrossover, "pcrossover")
  .chk_prob(x$pmutation, "pmutation")
  .chk_prob(x$pchangepoint, "pchangepoint")
  if (!.is_intish(x$minDist) || x$minDist < 0) stop("`minDist` must be integer >= 0.")
  if (!is.null(x$mmax) && (!.is_intish(x$mmax) || x$mmax < 0)) stop("`mmax` must be NULL or integer >= 0.")
  if (!is.null(x$lmax) && (!.is_intish(x$lmax) || x$lmax < 0)) stop("`lmax` must be NULL or integer >= 0.")
  if (!.is_intish(x$maxgen) || x$maxgen < 1) stop("`maxgen` must be integer >= 1.")
  if (!.is_intish(x$maxconv) || x$maxconv < 1) stop("`maxconv` must be integer >= 1.")
  if (!(x$option %in% c("cp", "both"))) stop("`option` must be 'cp' or 'both'.")
  if (!is.logical(x$monitoring) || length(x$monitoring) != 1L) stop("`monitoring` must be logical(1).")
  if (!is.logical(x$parallel) || length(x$parallel) != 1L) stop("`parallel` must be logical(1).")
  if (!is.null(x$nCore) && (!.is_intish(x$nCore) || x$nCore < 1)) stop("`nCore` must be NULL or integer >= 1.")
  if (!(is.numeric(x$tol) && length(x$tol) == 1L && is.finite(x$tol) && x$tol > 0)) stop("`tol` must be > 0.")
  if (!is.null(x$seed) && !.is_intish(x$seed)) stop("`seed` must be NULL or an integer.")
  if (!(is.character(x$popInitialize) || is.function(x$popInitialize))) stop("`popInitialize` must be name or function.")
  if (!(is.character(x$selection) || is.function(x$selection))) stop("`selection` must be name or function.")
  if (!(is.character(x$crossover) || is.function(x$crossover))) stop("`crossover` must be name or function.")
  if (!(is.character(x$mutation) || is.function(x$mutation))) stop("`mutation` must be name or function.")
  if (engine == "cptgaisl") {
    if (!.is_intish(x$numIslands) || x$numIslands < 1) stop("`numIslands` must be integer >= 1.")
    if (!.is_intish(x$maxMig) || x$maxMig < 0) stop("`maxMig` must be integer >= 0.")
  }
  x
}

#' Build Control List for \code{cptga}/\code{cptgaisl}
#'
#' @description
#' Convenience constructor for GA control parameters used by
#' \code{changepointGA::cptga} and \code{changepointGA::cptgaisl}. It merges
#' named overrides into engine-specific defaults
#' (\link{.cptga.default} or \link{.cptgaisl.default}), with light validation.
#'
#' @param ... Named overrides for control fields (e.g., \code{popSize},
#'   \code{pcrossover}, \code{minDist}, \code{numIslands}).
#' @param .list Optional named list of overrides (merged with \code{...}).
#' @param .persist Logical; if \code{TRUE}, persist updated defaults back into
#'   the target environment (not usually recommended in user code).
#' @param .env Environment where defaults live (defaults to \code{parent.frame()}).
#' @param .validate Logical; validate values/ranges (default \code{TRUE}).
#' @param engine Character; one of \code{"cptga"} or \code{"cptgaisl"} to select
#'   the default set and validation rules.
#'
#' @details
#' Unknown names are rejected. When both \code{...} and \code{.list} are present,
#' they are combined, with later entries overwriting earlier ones.
#'
#' @return A list of class \code{"cptgaControl"}.
#'
#' @seealso \link{gareg_knots}, \link{.cptga.default}, \link{.cptgaisl.default}
#' @export
cptgaControl <- function(..., .list = NULL, .persist = FALSE,
                         .env = asNamespace("GAReg"), .validate = TRUE,
                         engine = NULL) {
  overrides <- list(...)
  if (!is.null(.list)) {
    if (!is.list(.list)) stop("`.list` must be a named list.")
    overrides <- c(overrides, list(.list))
  }
  if (length(overrides) == 1L && is.list(overrides[[1]]) && !length(names(overrides))) {
    overrides <- overrides[[1]]
  }

  if (!is.null(engine)) {
    engine <- match.arg(engine, c("cptga", "cptgaisl"))
  } else {
    nm <- names(overrides)
    engine <- if (length(nm) && any(c("numIslands", "maxMig") %in% nm)) "cptgaisl" else "cptga"
  }

  defaults_name <- if (engine == "cptga") ".cptga.default" else ".cptgaisl.default"
  if (!exists(defaults_name, envir = .env, inherits = TRUE)) {
    stop(sprintf("`%s` not found in target environment.", defaults_name))
  }
  current <- get(defaults_name, envir = .env, inherits = TRUE)

  if (!length(overrides)) {
    return(structure(current, class = c("cptgaControl", "list")))
  }

  if (is.null(names(overrides)) || any(names(overrides) == "")) {
    stop("All control overrides must be *named*.")
  }

  unknown <- setdiff(names(overrides), names(current))
  if (length(unknown)) {
    stop("Unknown control parameter(s): ", paste(unknown, collapse = ", "))
  }

  updated <- current
  updated[names(overrides)] <- overrides
  if (.validate) updated <- .validate_ctrl(updated, engine = engine)

  if (.persist) {
    assign(defaults_name, updated, envir = .env)
    invisible(structure(updated, class = c("cptgaControl", "list")))
  } else {
    structure(updated, class = c("cptgaControl", "list"))
  }
}

#' Fixed-Knots Population Initializer
#'
#' @description
#' Initializes a population matrix for the fixed-knots GA. Each column is a
#' feasible chromosome sampled by \link{selectTau_uniform_exact}.
#'
#' @param popSize Integer; number of individuals (columns).
#' @param prange Optional hyperparameter range (unused here).
#' @param N Series length.
#' @param minDist Integer minimum spacing between adjacent changepoints.
#' @param Pb Unused placeholder (kept for compatibility).
#' @param mmax,lmax Integers; maximum number of knots and chromosome length.
#' @param fixedknots Integer; number of knots to place.
#'
#' @return Integer matrix of size \code{lmax x popSize}; each column is a
#'   chromosome \code{c(m, tau_1, ..., tau_m, N+1, ...)}.
#'
#' @seealso \link{selectTau_uniform_exact}, \link{gareg_knots}
#' @export
Popinitial_fixknots <- function(popSize, prange = NULL, N, minDist, Pb, mmax, lmax, fixedknots) {
  pop <- matrix(0, nrow = lmax, ncol = popSize)

  for (j in 1:popSize) {
    pop[, j] <- selectTau_uniform_exact(N, fixedknots, minDist, lmax)
  }

  return(pop)
}

#' Crossover Operator (Fixed-\eqn{m}) with Feasibility-First Restarts
#'
#' @description
#' Produces a child chromosome from two fixed-\eqn{m} parents (same number of
#' knots) by alternately sampling candidate knot locations from the parents and
#' enforcing the spacing constraint \code{diff(child) > minDist}. If a conflict
#' is encountered, the routine restarts the construction up to a small cap.
#'
#' @details
#' Let \code{mom} and \code{dad} be chromosomes of the form
#' \code{c(m, tau_1, ..., tau_m, ...)}. This operator:
#' \enumerate{
#'   \item Initializes an empty child of size \eqn{m}.
#'   \item Picks the first knot at random from \code{mom} or \code{dad}.
#'   \item For each subsequent position \eqn{i=2,\dots,m}, considers the
#'         pair \code{(mom[i], dad[i])} and chooses the first value that
#'         maintains the spacing constraint relative to the previously chosen
#'         knot (\code{> minDist}); if both work, one is chosen at random.
#'   \item If no feasible choice exists at some step, the construction restarts
#'         from the first position (up to a small cap governed internally by
#'         \code{up_tol}).
#' }
#' The result is written back as a full-length chromosome with the sentinel
#' \code{N+1} in position \code{m+2}, and zeros elsewhere.
#'
#' @param mom,dad Integer vectors encoding parent chromosomes:
#'   first entry \eqn{m} (number of changepoints), followed by \eqn{m} ordered
#'   knot locations.
#' @param prange Unused placeholder (kept for compatibility with other GA
#'   operators). Default \code{NULL}.
#' @param minDist Integer; minimum spacing between adjacent knots in the child.
#' @param lmax Integer; chromosome length (number of rows in the population
#'   matrix).
#' @param N Integer; series length. Used to place the sentinel \code{N+1} at
#'   position \code{m+2}.
#'
#' @return
#' An integer vector of length \code{lmax} encoding the child chromosome:
#' \code{c(m, child_knots, N+1, 0, 0, ...)}.
#'
#' @seealso
#' \link{crossover_fixknots}, \link{mutation_fixknots},
#' \link{selectTau_uniform_exact}, \link{Popinitial_fixknots},
#' \link{gareg_knots}
#'
#' @examples
#' \donttest{
#' N <- 120
#' lmax <- 30
#' minDist <- 5
#' m <- 3
#' mom <- c(m, c(20, 50, 90), rep(0, lmax - 1 - m))
#' mom[m + 2] <- N + 1
#' dad <- c(m, c(18, 55, 85), rep(0, lmax - 1 - m))
#' dad[m + 2] <- N + 1
#' child <- crossover_fixknots(mom, dad, minDist = minDist, lmax = lmax, N = N)
#' child
#' }
#'
#' @export
crossover_fixknots <- function(mom, dad, prange = NULL, minDist, lmax, N) {
  up_tol <- 30L
  max_restarts <- 50L

  output <- rep.int(0L, lmax)

  m_child <- as.integer(dad[1L])
  child <- rep.int(NA_integer_, m_child)

  if (all(mom[2:(m_child + 1)] == dad[2:(m_child + 1)])) {
    mom <- selectTau_uniform_exact(N, m_child, minDist, lmax)
  }

  mom_tau <- as.integer(mom[2L:(m_child + 1L)])
  dad_tau <- as.integer(dad[2L:(m_child + 1L)])
  co_tab <- as.integer(as.vector(rbind(mom_tau, dad_tau)))

  i <- 1L
  ii <- 0L
  restarts <- 0L
  tmppick <- NA_integer_
  while (i <= m_child) {
    if (i == 1L) {
      # first pick random from first pair of mom and dad
      tmppick <- sample(co_tab[1:2], size = 1L)
      child[1L] <- tmppick
      i <- i + 1L
    } else {
      # pick from the i^th pair of mom and dad (many cases)
      tmp_co_tab <- co_tab[((i - 1L) * 2L + 1L):(i * 2L)]
      ok <- which(tmp_co_tab - tmppick > minDist)
      if (length(ok) >= 1L) {
        tmppick <- if (length(ok) == 2L) sample(tmp_co_tab[ok], 1L) else tmp_co_tab[ok]
        child[i] <- tmppick

        if (i == m_child) {
          if (all(child == dad_tau) || all(child == mom_tau)) {
            if (ii > up_tol) break
            i <- 1L
            child[] <- NA_integer_
            restarts <- restarts + 1L
          } else {
            i <- i + 1L
          }
        } else {
          i <- i + 1L
        }
      } else {
        i <- 1L
        child[] <- NA_integer_
        restarts <- restarts + 1L
      }
    }

    ii <- ii + 1L
    if (restarts > max_restarts) {
      # select a new one to guaranteed progress
      child <- selectTau_uniform_exact(N, m_child, minDist, lmax)[2L:(m_child + 1L)]
      break
    }
  }

  output[1] <- m_child
  output[2:(m_child + 1)] <- child
  output[m_child + 2] <- N + 1

  return(output)
}

#' Exact Uniform Sampler of Feasible Changepoints
#'
#' @description
#' Samples \eqn{m} ordered changepoint indices uniformly from all feasible
#' configurations on \code{1:N} subject to a minimum spacing \code{minDist}.
#' Encodes the result as a chromosome for downstream GA operators.
#'
#' @param N Integer series length.
#' @param m Integer number of changepoints to place.
#' @param minDist Integer minimum spacing between adjacent changepoints.
#' @param lmax Integer chromosome length.
#'
#' @return Integer vector length \code{lmax}:
#'   \code{c(m, tau_1, ..., tau_m, N+1, 0, 0, ...)}.
#'
#' @seealso \link{Popinitial_fixknots}, \link{mutation_fixknots}
#' @export
selectTau_uniform_exact <- function(N, m, minDist, lmax) {
  output <- rep(0, lmax)

  if (N < (m + 1L) * minDist + 2L) {
    stop("Infeasible: need N >= (m+1)*minDist + 2.")
  }
  L0 <- 1L + minDist
  U0 <- N - m * minDist - 1L # (N - minDist - 1) - (m-1)*minDist
  S <- (U0 - L0 + 1L) + m - 1L
  if (S < m) {
    stop("Infeasible (S < m).")
  }
  picks <- sort(sample.int(S, m, replace = FALSE))
  h <- (L0 - 1L) + picks - (seq_len(m) - 1L)
  i <- h + (seq_len(m) - 1L) * minDist
  output[1] <- m
  output[2:(m + 1)] <- i
  output[m + 2] <- N + 1

  return(output)
}

#' Mutation Operator (Fixed-Knots)
#'
#' @description
#' Replaces a child with a fresh feasible sample having the same \eqn{m},
#' drawn by \link{selectTau_uniform_exact}.
#'
#' @param child Current chromosome (its first entry defines \eqn{m}).
#' @param p.range,Pb Unused placeholders (kept for compatibility).
#' @param minDist Integer minimum spacing.
#' @param lmax,mmax Integers; chromosome length and maximum \eqn{m} (unused).
#' @param N Integer series length.
#'
#' @return New feasible chromosome with the same \eqn{m}.
#'
#' @seealso \link{crossover_fixknots}
#' @export
mutation_fixknots <- function(child, p.range = NULL, minDist, Pb, lmax, mmax, N) {
  m <- child[1]
  childMut <- selectTau_uniform_exact(N, m, minDist, lmax)

  return(childMut)
}

#' Default Controls for \code{cptga}
#'
#' @description
#' Engine defaults used by \link{cptgaControl} when \code{engine = "cptga"}.
#' Not exported; shown here for reference.
#'
#' @format A named list with fields like \code{popSize}, \code{pcrossover},
#' \code{pmutation}, \code{pchangepoint}, \code{minDist}, \code{maxgen},
#' \code{option}, \code{selection}, \code{crossover}, \code{mutation}, etc.
#'
#' @seealso \link{cptgaControl}, \link{.cptgaisl.default}
#' @keywords internal
.cptga.default <- list(
  popSize = 200,
  pcrossover = 0.95,
  pmutation = 0.3,
  pchangepoint = 0.01,
  minDist = 1,
  mmax = NULL,
  lmax = NULL,
  maxgen = 50000,
  maxconv = 5000,
  option = "cp",
  monitoring = FALSE,
  parallel = FALSE,
  nCore = NULL,
  tol = 1e-05,
  seed = NULL,
  popInitialize = "random_population",
  suggestions = NULL,
  selection = "selection_linear_rank",
  crossover = "uniform_crossover",
  mutation = "mutation"
)

#' Default Controls for \code{cptgaisl} (Island GA)
#'
#' @description
#' Engine defaults used by \link{cptgaControl} when \code{engine = "cptgaisl"}.
#' Includes island-specific fields (e.g., \code{numIslands}, \code{maxMig}).
#'
#' @format A named list with fields like \code{popSize}, \code{numIslands},
#' \code{pcrossover}, \code{pmutation}, \code{maxMig}, \code{maxgen}, etc.
#'
#' @seealso \link{cptgaControl}, \link{.cptga.default}
#' @keywords internal
.cptgaisl.default <- list(
  popSize = 200,
  numIslands = 5,
  pcrossover = 0.95,
  pmutation = 0.3,
  pchangepoint = 0.01,
  minDist = 1,
  mmax = NULL,
  lmax = NULL,
  maxMig = 1000,
  maxgen = 50,
  maxconv = 100,
  option = "cp",
  monitoring = FALSE,
  parallel = FALSE,
  nCore = NULL,
  tol = 1e-05,
  seed = NULL,
  popInitialize = "random_population",
  suggestions = NULL,
  selection = "selection_linear_rank",
  crossover = "uniform_crossover",
  mutation = "mutation"
)



#' Truncated-power regression spline basis (internal)
#'
#' Builds a degree-\eqn{d} truncated-power (piecewise polynomial) basis:
#' \deqn{[1, x, x^2, \ldots, x^d, (x-k_1)_+^{d}, \ldots, (x-k_L)_+^{d}]}
#' where \eqn{(u)_+ = \max(u, 0)}.
#'
#' Knots are sorted and deduplicated internally. This helper does **not**
#' clip knots to the data range; use [splineX()] for a safer wrapper that
#' sanitizes knots.
#'
#' @param x Numeric vector of predictor values.
#' @param knots Numeric vector of interior knots.
#' @param degree Integer polynomial degree \eqn{d \ge 0}. Default: 3.
#' @param intercept Logical; include the intercept column \eqn{1}. Default: `TRUE`.
#'
#' @return A numeric matrix with \code{(d+1) + L} columns (with intercept),
#'   where \eqn{L} is the number of unique knots supplied. Columns are named
#'   \code{"x^0"}, \code{"x^1"}, … and \code{tp(<k>)^d}.
#'
#' @keywords internal
#' @noRd
tp_basis <- function(x, knots, degree = 3, intercept = TRUE) {
  x <- as.numeric(x)
  knots <- sort(unique(knots))
  d <- as.integer(degree)

  if (is.na(d) || d < 0) stop("`degree` must be a nonnegative integer.")

  # polynomial
  P <- outer(x, 0:d, `^`)
  if (!intercept) P <- if (d >= 1) P[, -1, drop = FALSE] else matrix(nrow = length(x), ncol = 0)
  colnames(P) <- if (intercept) paste0("x^", 0:d) else paste0("x^", 1:d)

  # truncated-power
  H <- if (length(knots)) outer(x, knots, function(xx, k) pmax(xx - k, 0)^d) else NULL
  if (!is.null(H)) colnames(H) <- paste0("tp(", signif(knots, 6), ")^", d)

  as.matrix(cbind(P, H))
}


#' Build spline design matrices (piecewise polynomials, natural cubic, B-spline)
#'
#' Unified wrapper to generate spline covariates for three common cases:
#' \itemize{
#'   \item \code{type = "ppolys"}: Degree-\eqn{d} regression spline via
#'         truncated-power **piecewise polynomials** (uses internal \code{tp_basis()}).
#'   \item \code{type = "ns"}: Degree-3 **natural cubic spline**; enforces
#'         \eqn{f''(a)=f''(b)=0} at the boundary.
#'   \item \code{type = "bs"}: Degree-\eqn{d} **B-spline** basis (unconstrained).
#' }
#'
#' Knots are sorted, no-duplicated, and any knots outside \code{range(x)} are
#' dropped with a warning. For \code{type = "ns"}, \code{degree} is ignored
#' (natural splines are cubic).
#'
#' @param x Numeric vector of predictor values.
#' @param knots Numeric vector of interior knots.
#' @param degree Integer polynomial degree for \code{"ppolys"} and \code{"bs"}.
#'   Ignored for \code{"ns"} (always cubic). Must be provided for
#'   \code{"ppolys"} and \code{"bs"}.
#' @param type One of \code{c("ppolys", "ns", "bs")}.
#' @param intercept Logical; include intercept column where applicable. Default: `TRUE`.
#'
#' @return A numeric design matrix. Attributes are attached:
#'   \itemize{
#'     \item \code{"knots"} — the interior knots used
#'     \item \code{"boundary"} — \code{range(x)}
#'     \item \code{"degree"} — effective degree (i.e., 3 for \code{"ns"})
#'     \item \code{"type"} — the requested spline type
#'   }
#'
#' @examples
#' set.seed(1)
#' x <- sort(rnorm(100))
#' k <- quantile(x, probs = c(.25, .5, .75))
#'
#' # 1) Piecewise polynomials (degree 3)
#' X_pp <- splineX(x, knots = k, degree = 3, type = "ppolys", intercept = TRUE)
#' dim(X_pp) # n x ((3+1) + 3) = n x 7
#'
#' # 2) Natural cubic spline (cubic, degree ignored)
#' X_ns <- splineX(x, knots = k, type = "ns", intercept = TRUE)
#'
#' # 3) B-spline basis (degree 3)
#' X_bs <- splineX(x, knots = k, degree = 3, type = "bs", intercept = TRUE)
#'
#' # Fit without a duplicated intercept:
#' # fit <- lm(y ~ 0 + X_pp)
#'
#' @export
#' @seealso \code{\link[splines]{bs}}, \code{\link[splines]{ns}}
splineX <- function(x, knots, degree = NULL,
                    type = c("ppolys", "ns", "bs"),
                    intercept = TRUE) {
  type <- match.arg(type)
  x <- as.numeric(x)
  rangeX <- range(x)

  # knots need to be unique, sorted, within [min(x), max(x)]
  knots <- sort(unique(knots))
  if (length(knots)) {
    inside <- knots >= rangeX[1] & knots <= rangeX[2]
    if (any(!inside)) {
      warning(
        "Dropping knots outside data range: ",
        paste(knots[!inside], collapse = ", ")
      )
      knots <- knots[inside]
    }
  }

  if (type %in% c("ppolys", "bs")) {
    if (is.null(degree)) stop("`degree` must be set for type='", type, "'.")
    d <- as.integer(degree)
    if (is.na(d) || d < 0) stop("`degree` must be a nonnegative integer.")
  } else {
    d <- NA_integer_
  }

  res <- switch(type,
    ppolys = tp_basis(x, knots = knots, degree = d, intercept = intercept),
    ns = splines::ns(x,
      knots = knots, Boundary.knots = rangeX,
      intercept = intercept
    ),
    bs = splines::bs(x,
      degree = d, knots = knots, Boundary.knots = rangeX,
      intercept = intercept
    )
  )

  attr(res, "knots") <- knots
  attr(res, "boundary") <- rangeX
  attr(res, "degree") <- if (!is.na(d)) d else 3L
  attr(res, "type") <- type
  res
}

#' S4 Class for Genetic Algorithm-Based Regression
#'
#' @name gareg-class
#' @title S4 Class Definition for `gareg`
#'
#' @slot call The matched call that created the object.
#' @slot N The effective size of the x grid used for knot search
#'   (i.e., `length(x_unique)`), typically the number of unique `x`.
#' @importClassesFrom changepointGA cptga cptgaisl
#' @importFrom methods show slotNames slot
NULL
setClassUnion("numericOrNULL", members = c("numeric", "NULL"))
setClassUnion("numericOrChara", members = c("numeric", "character"))
setClassUnion("listOrNULL", members = c("list", "NULL"))
setClassUnion("functionOrNULL", members = c("function", "NULL"))
setClassUnion("gaBackendORNULL", members = c("cptga", "cptgaisl", "ga", "gaisl", "NULL"))


#' GAReg result container
#'
#' @description
#' S4 container for GA-based regression/changepoint tasks. Holds the GA
#' backend fit and a normalized summary of the best solution.
#'
#' @slot call language. The original call.
#' @slot method character. One of "varyknots", "fixknots", "subset".
#' @slot N numeric. Length of `x_unique` used by the GA (also `sentinel-1`).
#' @slot objFunc functionOrNULL. Objective function used.
#' @slot gaMethod character. GA engine name ("cptga","cptgaisl","ga","gaisl").
#' @slot gaFit Backend GA fit object (union of classes from GA and changepointGA).
#' @slot ctrl listOrNULL. Control list used to run the GA (if stored by caller).
#' @slot fixedknots numericOrNULL. Fixed number of interior knots (`m`) for fixed-knots mode, or NULL.
#' @slot minDist numeric. Minimum distance between adjacent changepoints.
#' @slot polydegree numericOrNULL. Spline degree for default objectives.
#' @slot type character. One of `c("ppolys", "ns", "bs")` indicating piecewise
#'       polynomials, natural cubic, or B-spline.
#' @slot intercept logical. Whether the spline basis included an intercept column.
#' @slot subsetSpec listOrNULL. Constraints for subset selection (unused for knots).
#' @slot featureNames character. Candidate feature names (subset tasks).
#' @slot bestFitness numeric. Best fitness value found.
#' @slot bestChrom numeric. Raw best chromosome returned by the backend
#'       (may include a sentinel equal to `N+1` and optional padding).
#' @slot bestnumbsol numeric. Count of selected elements (e.g., `m` for knots).
#' @slot bestsol numericOrChara. For knots: the `m` interior indices (pre-sentinel);
#'       for subset: mask/indices/names.
#' @seealso [gareg_knots], [cptgaControl]
#' @exportClass gareg
setClass(
  Class = "gareg",
  slots = c(
    call = "language",
    method = "character",
    N = "numeric",
    objFunc = "functionOrNULL",
    gaMethod = "character",
    gaFit = "gaBackendORNULL",
    ctrl = "listOrNULL",
    # knots
    fixedknots = "numericOrNULL",
    minDist = "numeric",
    polydegree = "numericOrNULL",
    type = "character",
    intercept = "logical",
    # best subset
    subsetSpec = "listOrNULL",
    featureNames = "character",
    # general results
    bestFitness = "numeric",
    bestChrom = "numeric",
    bestnumbsol = "numeric",
    bestsol = "numericOrChara"
  ),
  prototype = list(
    method       = character(),
    fixedknots   = NA_real_,
    minDist      = numeric(),
    polydegree   = NA_real_,
    type         = character(),
    intercept    = logical(),
    subsetSpec   = NULL,
    featureNames = character(),
    ctrl         = NULL,
    bestFitness  = numeric(),
    bestChrom    = numeric(),
    bestnumbsol  = numeric(),
    bestsol      = numeric()
  ),
  package = "GAReg"
)


.s <- function(x, nm, default = NA) {
  if (nm %in% methods::slotNames(x)) methods::slot(x, nm) else default
}


#' @export
setMethod("show", "gareg", function(object) {
  hdr <- switch(object@method,
    subset    = "# Best Subset Variable Selection via GA                  #",
    varyknots = "# Varying-Knots Spline Regression via changepointGA     #",
    fixknots  = "# Fixed-Knots Spline Regression via changepointGA       #",
    "# GAReg Result                                                 #"
  )

  cat("##########################################################\n")
  cat(hdr, "\n", sep = "")
  cat("##########################################################\n")
  cat("Call: ")
  print(object@call)
  cat("   gaMethod: ", object@gaMethod, "\n", sep = "")
  cat("   N (|x_unique|): ", object@N, "\n", sep = "")
  if (object@method %in% c("varyknots", "fixknots")) {
    cat("   Spline type / degree / intercept: ",
      paste0(object@type, " / ", object@polydegree, " / ", object@intercept), "\n",
      sep = ""
    )
    cat("   minDist: ", object@minDist, "\n", sep = "")
  }
  if (!is.null(object@gaFit)) cat("\nUse summary() for GA settings and best solution.\n")
})

# internal pretty-printer used by summary()
#' @keywords internal
#' @noRd
print_summary_gareg <- function(x, digits = getOption("digits"), max_display = 5, ...) {
  gf <- x@gaFit
  cat("##########################################################\n")
  hdr <- switch(x@method,
    subset    = "# Best Subset Variable Selection via GA                  #",
    varyknots = "# Varying-Knots Spline Regression via changepointGA     #",
    fixknots  = "# Fixed-Knots Spline Regression via changepointGA       #",
    "# GAReg Result                                                 #"
  )
  cat(hdr, "\n", sep = "")
  cat("##########################################################\n")

  # --- GA hyperparameters ---
  cat("   Settings: \n")
  cat("   Population size         = ", .s(gf, "popSize", "NA"), "\n", sep = "")
  cat("   Number of generations   = ", .s(gf, "count", "NA"), "\n", sep = "")
  cat("   GA convergence          = ", .s(gf, "convg", FALSE), "\n", sep = "")
  cat("   Crossover probability   = ", format(.s(gf, "pcrossover", NA), digits = digits), "\n", sep = "")
  cat("   Mutation probability    = ", format(.s(gf, "pmutation", NA), digits = digits), "\n", sep = "")
  cat("   Changepoint probability = ", format(.s(gf, "pchangepoint", NA), digits = digits), "\n", sep = "")
  cat("   Parallel Usage          = ", .s(gf, "parallel", FALSE), "\n", sep = "")
  if (isTRUE(.s(gf, "parallel", FALSE))) {
    cat("   Number of threads       = ", .s(gf, "nCore", "NA"), "\n", sep = "")
  }
  sugg <- .s(gf, "suggestions", NULL)
  if (!is.null(sugg)) {
    cat("   Suggestions:\n")
    for (i in seq_along(sugg)) {
      cat("     [", i, "]: ", paste(sugg[[i]], collapse = " "), "\n", sep = "")
      if (i >= max_display) {
        cat("     ...\n")
        break
      }
    }
  }

  # --- Best solution ---
  cat("\n##### Best ##### \n")
  m <- x@bestnumbsol
  bs <- x@bestsol

  if (length(bs) == 0L || all(is.na(bs))) {
    chrom <- x@bestChrom
    if (length(chrom)) {
      # decode from chromosome
      m_int <- as.integer(chrom[1L])
      Nu <- as.integer(x@N)
      tail <- as.integer(chrom[-1L])
      endp <- match(Nu + 1L, tail) # sentinel = N+1 (length(x_unique)+1)
      idx <- if (!is.na(endp)) tail[seq_len(endp - 1L)] else integer()
      idx <- idx[idx != 0L]
      if (!is.na(m_int) && length(idx) >= m_int) {
        idx <- idx[seq_len(m_int)]
      }
      bs <- idx
      m <- m_int
    }
  }

  if (length(bs) == 0L || all(is.na(bs))) {
    cat("   <no best solution stored>\n")
  } else if (x@method %in% c("varyknots", "fixknots", "cptdetect")) {
    ## Knots-style: bestnumbsol = m, bestsol = indices into x_unique
    m_int <- as.integer(if (length(m)) m[1] else NA_integer_)
    tau <- as.integer(bs)
    if (!is.na(m_int) && length(tau) >= m_int) {
      tau <- tau[seq_len(m_int)]
    }
    cat("   Fitness   = ", format(x@bestFitness, digits = digits), "\n", sep = "")
    cat("   m         = ", if (is.na(m_int)) 0L else m_int, "\n", sep = "")
    cat("   indices   = ", if (length(tau)) paste(tau, collapse = " ") else "<none>", "\n", sep = "")
    cat("   sentinel  = N+1 = ", as.integer(x@N) + 1L, "\n", sep = "")
  } else if (identical(x@method, "subset")) {
    ## Subset-GA style: bestsol is either a 0/1 mask, indices, or names
    fn <- x@featureNames

    if (is.numeric(bs) && length(fn) > 0L && length(bs) == length(fn) && all(bs %in% c(0, 1))) {
      # 0/1 mask
      idx <- which(bs != 0)
      k <- length(idx)
      nm <- if (k) fn[idx] else character(0)
      cat("   Fitness = ", format(x@bestFitness, digits = digits), "\n", sep = "")
      cat("   k       = ", k, "\n", sep = "")
      cat("   subset  = ", if (k) paste(nm, collapse = ", ") else "<none>", "\n", sep = "")
    } else if (is.numeric(bs)) {
      # integer indices
      idx <- sort(unique(as.integer(bs[is.finite(bs) & bs > 0])))
      k <- length(idx)
      nm <- if (length(fn) >= max(c(idx, 0))) fn[idx] else as.character(idx)
      cat("   Fitness    = ", format(x@bestFitness, digits = digits), "\n", sep = "")
      cat("   k          = ", k, "\n", sep = "")
      cat("   Subset Id  = ", paste(format(x@bestsol, digits = digits), collapse = " "), "\n", sep = "")
      cat("   Best subset= ", if (k) paste(nm, collapse = ", ") else "<none>", "\n", sep = "")
    } else if (is.character(bs)) {
      # feature names directly
      nm <- unique(bs[nzchar(bs)])
      k <- length(nm)
      cat("   Fitness = ", format(x@bestFitness, digits = digits), "\n", sep = "")
      cat("   k       = ", k, "\n", sep = "")
      cat("   subset  = ", if (k) paste(nm, collapse = ", ") else "<none>", "\n", sep = "")
    } else {
      cat("   Fitness = ", format(x@bestFitness, digits = digits), "\n", sep = "")
      cat("   bestsol = ", paste(bs, collapse = " "), "\n", sep = "")
    }
  } else {
    ## Fallback: just show the vector
    cat("   Fitness = ", format(x@bestFitness, digits = digits), "\n", sep = "")
    cat("   bestsol = ", paste(bs, collapse = " "), "\n", sep = "")
  }

  invisible(x)
}

#' @export
setMethod("summary", "gareg", function(object, ...) print_summary_gareg(object, ...))

#' Genetic-Algorithm Best Subset Selection (GA / GAISL)
#'
#' @description
#' Runs a GA-based search over variable subsets using a user-specified
#' objective (default: \code{\link{subsetBIC}}) and returns a compact
#' \code{"gareg"} S4 result with \code{method = "subset"}.
#' The engine can be \code{\link[GA]{ga}} (single population) or
#' \code{\link[GA]{gaisl}} (islands), selected via \code{gaMethod}.
#'
#' @param y Numeric response vector (length \code{n}).
#' @param X Numeric matrix of candidate predictors (\code{n} rows by \code{p} columns).
#' @param ObjFunc Objective function or its name. Defaults to \code{\link{subsetBIC}}.
#'   The objective must accept as its first argument a binary chromosome
#'   (0/1 mask of length \code{p}) and may accept additional arguments passed via \code{...}.
#'   By convention, \code{subsetBIC} returns \emph{negative} BIC, so the GA maximizes fitness.
#' @param gaMethod GA backend to call: \code{"ga"} or \code{"gaisl"} (functions from package \pkg{GA}),
#'   or a GA-compatible function with the same interface as \code{\link[GA]{ga}}.
#' @param gacontrol Optional named list of GA engine controls (e.g., \code{popSize}, \code{maxiter},
#'   \code{run}, \code{pcrossover}, \code{pmutation}, \code{elitism}, \code{seed}, \code{parallel},
#'   \code{keepBest}, \code{monitor}, ...). These are passed to the GA engine, not to the objective.
#' @param monitoring Logical; if \code{TRUE}, prints a short message and (if not supplied in
#'   \code{gacontrol}) sets \code{monitor = GA::gaMonitor} for live progress.
#' @param seed Optional RNG seed (convenience alias for \code{gacontrol$seed}).
#' @param ... Additional arguments forwarded to \code{ObjFunc} (not to the GA engine). For
#'   \code{\link{subsetBIC}} these typically include \code{family}, \code{weights}, \code{offset},
#'   and \code{control}.
#'
#' @details
#' The fitness passed to \pkg{GA} is \code{ObjFunc} itself. Because the engine expects
#' a function with signature \code{f(chrom, ...)}, your \code{ObjFunc} must interpret
#' \code{chrom} as a 0/1 mask over the columns of \code{X}. The function then computes a score
#' (e.g., negative BIC) using \code{y}, \code{X}, and any extra arguments supplied via \code{...}.
#'
#' With the default \code{\link{subsetBIC}}, the returned value is \code{-BIC}, so we set
#' \code{max = TRUE} in the GA call to maximize fitness. If you switch to an objective that
#' returns a quantity to \emph{minimize}, either negate it in your objective or change
#' the engine setting to \code{max = FALSE}.
#'
#' Engine controls belong in \code{gacontrol}; objective-specific options belong in \code{...}.
#' This separation prevents accidental name collisions between GA engine parameters and
#' objective arguments.
#'
#' @return An object of S4 class \code{"gareg"} (with \code{method = "subset"}) containing:
#' \itemize{
#'   \item \code{call} – the matched call.
#'   \item \code{N} – number of observations.
#'   \item \code{objFunc} – the objective function used.
#'   \item \code{gaMethod} – \code{"ga"} or \code{"gaisl"}.
#'   \item \code{gaFit} – the GA fit object returned by \pkg{GA} (if your class allows it).
#'   \item \code{featureNames} – column names of \code{X} (or empty).
#'   \item \code{bestFitness} – best fitness value (\code{GA::ga}@fitnessValue).
#'   \item \code{bestChrom} – \code{c(m, idx)}: number of selected variables and their indices.
#'   \item \code{bestnumbsol} – \code{m}, number of selected variables.
#'   \item \code{bestsol} – vector of selected column indices in \code{X}.
#' }
#'
#' @seealso
#' \code{\link{subsetBIC}},
#' \code{\link[GA]{ga}},
#' \code{\link[GA]{gaisl}}
#'
#' @examples
#' \donttest{
#' if (requireNamespace("GA", quietly = TRUE)) {
#'   set.seed(1)
#'   n <- 100
#'   p <- 12
#'   X <- matrix(rnorm(n * p), n, p)
#'   y <- 1 + X[, 1] - 0.7 * X[, 4] + rnorm(n, sd = 0.5)
#'
#'   # Default: subsetBIC (Gaussian – negative BIC), engine = GA::ga
#'   fit1 <- gareg_subset(y, X,
#'     gaMethod = "ga",
#'     gacontrol = list(popSize = 60, maxiter = 80, run = 40, parallel = FALSE)
#'   )
#'   summary(fit1)
#'
#'   # Island model: GA::gaisl
#'   fit2 <- gareg_subset(y, X,
#'     gaMethod = "gaisl",
#'     gacontrol = list(popSize = 40, maxiter = 60, numIslands = 4, parallel = FALSE)
#'   )
#'   summary(fit2)
#'
#'   # Logistic objective (subsetBIC handles GLM via ...):
#'   ybin <- rbinom(n, 1, plogis(0.3 + X[, 1] - 0.5 * X[, 2]))
#'   fit3 <- gareg_subset(ybin, X,
#'     gaMethod = "ga",
#'     family = stats::binomial(), # <- passed to subsetBIC via ...
#'     gacontrol = list(popSize = 60, maxiter = 80, parallel = FALSE)
#'   )
#'   summary(fit3)
#' }
#' }
#'
#' @export
gareg_subset <- function(y,
                         X,
                         ObjFunc = NULL,
                         gaMethod = "ga",
                         gacontrol = NULL,
                         monitoring = FALSE,
                         seed = NULL,
                         ...) {
  if (!requireNamespace("GA", quietly = TRUE)) {
    stop("Package 'GA' is required: install.packages('GA')")
  }

  call <- match.call()
  X <- as.matrix(X)
  nobs <- NROW(X)
  p <- NCOL(X)

  ga_fun <- if (is.function(gaMethod)) {
    gaMethod
  } else {
    switch(tolower(as.character(gaMethod)),
      "ga"    = GA::ga,
      "gaisl" = GA::gaisl,
      stop("gaMethod must be 'ga' or 'gaisl' (from the GA package), or a GA-compatible function.")
    )
  }
  ga_name <- if (identical(ga_fun, GA::gaisl)) "gaisl" else "ga"

  if (is.null(ObjFunc)) {
    ObjFunc <- subsetBIC
  } else if (is.character(ObjFunc)) {
    ObjFunc <- get(ObjFunc, mode = "function")
  }
  if (!is.function(ObjFunc)) stop("ObjFunc must be a function or a function name.", call. = FALSE)

  engine_args <- as.list(gacontrol %||% list())
  if (!is.null(seed)) engine_args$seed <- seed
  if (isTRUE(monitoring) && is.null(engine_args$monitor)) engine_args$monitor <- GA::gaMonitor

  if (isTRUE(monitoring)) message("Running best subset via GA (engine = ", ga_name, ")")

  base_args <- list(
    fitness = ObjFunc, # default as subsetBIC; expect chromosome with 0/1 mask
    type    = "binary",
    nBits   = p,
    monitor = monitoring
  )

  GA.res <- do.call(
    ga_fun,
    c(
      base_args,
      engine_args,
      list(y = y, X = X),
      list(...)
    )
  )

  sol <- GA.res@solution
  if (is.matrix(sol)) sol <- sol[1L, ]
  idx <- which(sol != 0)
  mhat <- length(idx)
  bestChrom <- c(mhat, idx)

  `%||%` <- function(a, b) if (!is.null(a)) a else b
  feat_names <- colnames(X) %||% character(p)

  # Build gareg S4
  object <- methods::new("gareg",
    call         = call,
    method       = "subset",
    N            = nobs,
    objFunc      = ObjFunc,
    gaMethod     = ga_name,
    gaFit        = GA.res,
    featureNames = feat_names,
    bestFitness  = as.numeric(GA.res@fitnessValue),
    bestChrom    = as.numeric(bestChrom),
    bestnumbsol  = as.numeric(mhat),
    bestsol      = as.numeric(idx)
  )

  return(object)
}

#' Unified BIC-style Objective for Subset Selection (GLM & Gaussian)
#'
#' @description
#' Computes a BIC-like criterion for a chromosome that encodes a variable
#' subset. The same expression
#' \deqn{\mathrm{BIC} = n \log(\mathrm{rss\_like}/n) + k \log n}
#' is used for all families, where:
#' \itemize{
#'   \item For Gaussian with identity link, \code{rss_like} is the residual sum of squares (RSS),
#'         computed via a fast \code{.lm.fit}.
#'   \item For other GLM families, \code{rss_like} is the residual \emph{deviance}
#'         from \code{glm.fit}.
#' }
#' The effective parameter count \eqn{k} includes the intercept.
#'
#' @details
#' The chromosome \code{subset_bin} is a \emph{binary} vector (0/1 by column),
#' indicating which predictors from \code{X} are included. The design matrix
#' always includes an intercept. Rank-deficient selections return \code{Inf}
#' (which the GA maximizer treats as a very poor score). The value returned is
#' \strong{-BIC} so that GA engines can \emph{maximize} it.
#'
#' @param subset_bin Integer/numeric 0–1 vector (length \code{ncol(X)}); 1 means
#'   the corresponding column of \code{X} is included in the model.
#' @param y Numeric response vector of length \code{n}.
#' @param X Numeric matrix of candidate predictors; columns correspond to variables.
#' @param family A GLM family object (default \code{stats::gaussian()}).
#' @param weights Optional prior weights (passed to \code{glm.fit}).
#' @param offset Optional offset (passed to \code{glm.fit}).
#' @param control GLM fit controls; default \code{stats::glm.control()}.
#'
#' @return A single numeric value: \strong{-BIC}. Larger is better for GA maximizers.
#'   Returns \code{Inf} for rank-deficient designs.
#'
#' @seealso \code{\link[stats]{glm.fit}}, \code{\link[stats]{glm.control}},
#'   \code{\link[stats]{.lm.fit}}
#'
#' @export
#' @importFrom stats glm.fit glm.control .lm.fit
subsetBIC <- function(subset_bin,
                      y,
                      X,
                      family = stats::gaussian(),
                      weights = NULL,
                      offset = NULL,
                      control = stats::glm.control()) {
  n <- as.integer(length(y))
  X <- as.matrix(X)

  idX <- which(as.integer(abs(subset_bin) != 0) == 1L)
  m <- length(idX)

  ## Construct the Design Matrix
  if (m == 0L) {
    Xmat <- matrix(1,
      nrow = n, ncol = 1L,
      dimnames = list(NULL, "(Intercept)")
    )
  } else {
    Xsel <- X[, idX, drop = FALSE]
    Xmat <- cbind(`(Intercept)` = 1, Xsel)
    if (qr(Xmat)$rank < ncol(Xmat)) BIC_val <- Inf
  }

  k_eff <- NCOL(Xmat) # include intercept

  if (identical(family$family, "gaussian") && identical(family$link, "identity")) {
    ## Gaussian-identity: use RSS
    fit <- stats::.lm.fit(Xmat, y)
    if (!is.null(fit$rank) && fit$rank < k_eff) BIC_val <- Inf
    rss_like <- sum(fit$residuals^2)
  } else {
    ## GLM: use deviance
    fit <- stats::glm.fit(
      x = Xmat, y = y,
      family = family,
      weights = weights,
      offset = offset,
      control = control
    )
    if (!is.null(fit$rank) && fit$rank < k_eff) BIC_val <- Inf
    rss_like <- fit$deviance
    if (!is.finite(rss_like) || rss_like <= 0) {
      # avoid -Inf BIC
      rss_like <- .Machine$double.eps
    }
  }

  BIC_val <- n * log(rss_like / n) + k_eff * log(n)

  return(-BIC_val) # GA::ga or GA::gaisl maximizing
}



#' False Discovery Rate (FDR) and True Positive Rate (TPR) from index labels
#'
#' @description
#' Computes the False Discovery Rate (FDR) and True Positive Rate (TPR, a.k.a. recall)
#' by comparing a set of \emph{true} labels to a set of \emph{predicted} labels.
#' Labels are treated as positive integer indices in \eqn{\{1, \dots, N\}}. Duplicates
#' are ignored (unique indices are used).
#'
#' @details
#' Let \code{truelabel} and \code{predlabel} be sets of indices. The function
#' derives the confusion-matrix counts:
#' \itemize{
#'   \item \code{tp} = \eqn{|truelabel \cap predlabel|}
#'   \item \code{fp} = \eqn{|predlabel \setminus truelabel|}
#'   \item \code{fn} = \eqn{|truelabel \setminus predlabel|}
#'   \item \code{tn} = \eqn{N - tp - fp - fn}
#' }
#' and returns
#' \deqn{FDR = fp / (fp + tp), \qquad TPR = tp / (tp + fn).}
#'
#' Inputs are coerced to integer and uniqued. A warning is emitted if any label
#' is < 1, and an error is thrown if any label exceeds \code{N}. If \code{tn < 0},
#' a warning is issued indicating that \code{N} may not reflect the full universe.
#'
#' @param truelabel Integer vector of ground-truth positive indices (values in \code{1..N}).
#' @param predlabel Integer vector of predicted positive indices (values in \code{1..N}).
#' @param N Integer scalar; size of the full index universe (total number of candidates).
#'
#' @return A named list with components:
#' \itemize{
#'   \item \code{fdr} False Discovery Rate, \eqn{fp/(fp+tp)} (NaN if \code{fp+tp == 0}).
#'   \item \code{tpr} True Positive Rate (recall), \eqn{tp/(tp+fn)} (NaN if \code{tp+fn == 0}).
#'   \item \code{fp}, \code{fn}, \code{tp}, \code{tn} Confusion-matrix counts.
#' }
#'
#' @section Edge cases:
#' If \code{predlabel} is empty, \code{fdr} is \code{NaN} and \code{tpr} is 0 (unless
#' \code{truelabel} is also empty, in which case both \code{fdr} and \code{tpr} are \code{NaN}).
#' If \code{truelabel} is empty and \code{predlabel} non-empty, \code{tpr} is \code{NaN}
#' and \code{fdr} is 1.
#'
#' @examples
#' # Simple example
#' N <- 10
#' true <- c(2, 4, 7)
#' pred <- c(4, 5, 7, 7) # duplicates are ignored
#' FDRCalc(true, pred, N)
#'
#' # Empty predictions
#' FDRCalc(true, integer(0), N)
#'
#' # All correct predictions
#' FDRCalc(true, true, N)
#'
#' @export
FDRCalc <- function(truelabel, predlabel, N) {
  truelabel <- unique(as.integer(truelabel))
  predlabel <- unique(as.integer(predlabel))

  tp <- length(intersect(truelabel, predlabel))
  fp <- length(setdiff(predlabel, truelabel))
  fn <- length(setdiff(truelabel, predlabel))

  if (any(truelabel < 1L) | any(predlabel < 1L)) {
    warning("Labels are expected to be positive indices (>= 1).")
  }
  if (any(truelabel > N) | any(predlabel > N)) {
    stop("Found labels greater than N. Increase N or correct labels.")
  }


  tn <- N - tp - fp - fn
  if (tn < 0) warning("Computed TN < 0; check that N reflects the full universe.")

  fdr <- fp / (fp + tp)
  tpr <- tp / (tp + fn)

  return(list(fdr = fdr, tpr = tpr, fp = fp, fn = fn, tp = tp, tn = tn))
}

#' Methods for displaying and summarizing `gareg` objects
#'
#' @title Show and summary methods for \code{gareg}
#' @description
#' \itemize{
#'   \item \code{show(object)}: Compact header with call, engine, and N.
#'   \item \code{summary(object, ...)}: GA settings (when available) and best solution.
#' }
#' @param object A \code{"gareg"} object.
#' @param ... Currently unused.
#' @return \code{show}: invisible NULL. \code{summary}: invisibly returns \code{object}.
#' @seealso \link{gareg-class}
#' @name gareg-methods
NULL

#' @keywords internal
#' @noRd
if (!isGeneric("summary")) {
  setGeneric("summary", function(object, ...) standardGeneric("summary"))
}

#' @rdname gareg-methods
#' @export
setMethod("show", "gareg", function(object) {
  hdr <- switch(object@method,
    subset    = "# Best Subset Variable Selection via GA                  #",
    varyknots = "# Varying Knots Detection via changepointGA              #",
    fixknots  = "# Fixed Knots Detection via changepointGA                #",
    "# GAReg Result                                                 #"
  )
  cat("##########################################################\n")
  cat(hdr, "\n", sep = "")
  cat("##########################################################\n")
  cat("Call: ")
  print(object@call)
  cat("   gaMethod: ", object@gaMethod, "\n", sep = "")
  cat("N: ", object@N, "\n", sep = "")
  if (!is.null(object@gaFit)) cat("\nUse summary() for GA settings and best solution.\n")
  invisible(NULL)
})

#' @keywords internal
#' @noRd
print_summary_gareg <- function(x, digits = getOption("digits"), max_display = 5, ...) {
  gf <- x@gaFit
  cat("##########################################################\n")
  hdr <- switch(x@method,
    subset    = "# Best Subset Variable Selection via GA                  #",
    varyknots = "# Varying Knots Detection via changepointGA              #",
    fixknots  = "# Fixed Knots Detection via changepointGA                #",
    "# GAReg Result                                                 #"
  )
  cat(hdr, "\n", sep = "")
  cat("##########################################################\n")
  cat("   Settings: \n")
  cat("   Population size         = ", .s(gf, "popSize", "NA"), "\n", sep = "")
  cat("   Number of generations   = ", .s(gf, "count", "NA"), "\n", sep = "")
  cat("   GA convergence          = ", .s(gf, "convg", FALSE), "\n", sep = "")
  cat("   Crossover probability   = ", format(.s(gf, "pcrossover", NA), digits = digits), "\n", sep = "")
  cat("   Mutation probability    = ", format(.s(gf, "pmutation", NA), digits = digits), "\n", sep = "")
  cat("   Changepoint probability = ", format(.s(gf, "pchangepoint", NA), digits = digits), "\n", sep = "")
  cat("   Parallel Usage          = ", .s(gf, "parallel", FALSE), "\n", sep = "")
  if (isTRUE(.s(gf, "parallel", FALSE))) {
    cat("   Number of thread        = ", .s(gf, "nCore", "NA"), "\n", sep = "")
  }
  sugg <- .s(gf, "suggestions", NULL)
  if (!is.null(sugg)) {
    cat("   Suggestions:\n")
    for (i in seq_along(sugg)) {
      cat("     [", i, "]: ", paste(sugg[[i]], collapse = " "), "\n", sep = "")
      if (i >= max_display) {
        cat("     ...\n")
        break
      }
    }
  }

  cat("\n##### Best ##### \n")
  m <- x@bestnumbsol
  bs <- x@bestsol

  if (length(bs) == 0L || all(is.na(bs))) {
    cat("   <no best solution stored>\n")
  } else if (x@method %in% c("varyknots", "fixknots", "cptdetect")) {
    m_int <- as.integer(if (length(m)) m[1] else NA_integer_)
    tau <- as.integer(bs)
    if (!is.na(m_int) && length(tau) >= m_int) tau <- tau[seq_len(m_int)]
    cat("   Fitness   =", format(x@bestFitness, digits = digits), "\n")
    cat("   m         =", if (is.na(m_int)) 0L else m_int, "\n")
    cat("   knots     =", if (length(tau)) paste(tau, collapse = " ") else "<none>", "\n")
  } else if (identical(x@method, "subset")) {
    fn <- x@featureNames
    if (is.numeric(bs) && length(fn) > 0L && length(bs) == length(fn) && all(bs %in% c(0, 1))) {
      idx <- which(bs != 0)
      k <- length(idx)
      nm <- if (k) fn[idx] else character(0)
      cat("   Fitness =", format(x@bestFitness, digits = digits), "\n")
      cat("   k =", k, "\n")
      cat("   subset =", if (k) paste(nm, collapse = ", ") else "<none>", "\n")
    } else if (is.numeric(bs)) {
      idx <- sort(unique(as.integer(bs[is.finite(bs) & bs > 0])))
      k <- length(idx)
      nm <- if (length(fn) >= max(c(idx, 0))) fn[idx] else as.character(idx)
      cat("   Fitness =", format(x@bestFitness, digits = digits), "\n")
      cat("   k =", k, "\n")
      cat("   Subset Id =", format(x@bestsol, digits = digits), "\n")
      cat("   Best subset =", if (k) paste(nm, collapse = ", ") else "<none>", "\n")
    } else if (is.character(bs)) {
      nm <- unique(bs[nzchar(bs)])
      k <- length(nm)
      cat("   Fitness =", format(x@bestFitness, digits = digits), "\n")
      cat("   k =", k, "\n")
      cat("   subset =", if (k) paste(nm, collapse = ", ") else "<none>", "\n")
    } else {
      cat("   Fitness =", format(x@bestFitness, digits = digits), "\n")
      cat("   bestsol =", paste(bs, collapse = " "), "\n")
    }
  } else {
    cat("   Fitness =", format(x@bestFitness, digits = digits), "\n")
    cat("   bestsol =", paste(bs, collapse = " "), "\n")
  }

  invisible(x)
}

#' @rdname gareg-methods
#' @export
setMethod("summary", "gareg", function(object, ...) {
  print_summary_gareg(object, ...)
  invisible(object)
})

---
title: "A brief guide to GAReg"
author: "Mo Li, QiQi Lu, Robert Lund, Xueheng Shi"
date: "`r format(Sys.time(), '%d %b %Y')`"
output: 
  rmarkdown::html_vignette:
    number_sections: false
    toc: true
vignette: >
  %\VignetteIndexEntry{A brief guide to GAReg}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE
)

set.seed(12345)
```


## Overview

**GAreg** implements genetic algorithm (GA) optimization for three core statistical workflows where discrete model structure must be learned from data:

1. Best subset variable selection – choose a parsimonious set of predictors under information or validation criteria.

2. Changepoint detection – estimate the number and locations of structural breaks in regression or time series.

3. Optimal spline knot selection – place interior knots for splines or piecewise polynomials to balance fit and smoothness.

The package provides a consistent GA interface, unified S4 results (gareg), and penalty-aware objectives. It supports both varying‑knot and fixed‑knot modes, minimum spacing constraints, and unbalanced designs.

```{r setup}
library(GAReg)
```

## Best subset variable selection

We use the simulation function below for subset selection illustration. Here, `n` is the number of observations and `p` is the number of predictors. For the covairates, `s0` represent number of truly active predictors, valued range from `0` to `p`. `magnitudes_range` specifies the range of significantly expressed coefficients that corresponding to the truly active predictors. If `rho` is specified with some values, the autoregressive structure is introduced into the error terms. If `rho=NULL`, we will have independent and identically distributed (IID) errors. We can also specify `sigma` for the error standard deviation.

```{r}
sim_subset_data <- function(n = 60, p = 50, s0 = 25, sigma = 1.5,
                            magnitudes_range = c(0.5, 2),
                            rho = NULL,
                            seed = NULL) {
  stopifnot(n > 0, p > 0, s0 >= 0, s0 <= p, sigma >= 0)
  if (!is.null(seed)) set.seed(seed)

  X <- matrix(rnorm(n * p), n, p)

  # Active set and coefficients
  true_idx <- if (s0 > 0) sort(sample.int(p, s0)) else integer(0)
  signs <- if (s0 > 0) sample(c(-1, 1), s0, replace = TRUE) else numeric(0)
  magnitudes <- if (s0 > 0) runif(s0, magnitudes_range[1], magnitudes_range[2]) else numeric(0)

  beta_true <- numeric(p)
  if (s0 > 0) beta_true[true_idx] <- magnitudes * signs

  if (is.null(rho)) {
    e <- rnorm(n, sd = sigma)
  } else {
    sd_innov <- sigma * sqrt(1 - rho^2)
    burn_in <- 100
    z <- rnorm(n + burn_in, sd = sd_innov)
    e_full <- numeric(n + burn_in)
    for (t in 2:(n + burn_in)) e_full[t] <- rho * e_full[t - 1] + z[t]
    e <- e_full[(burn_in + 1):(burn_in + n)]
  }

  y <- as.numeric(X %*% beta_true + e)

  DF <- data.frame(y = y, as.data.frame(X))
  colnames(DF)[-1] <- paste0("X", seq_len(p))

  list(
    X = X,
    y = y,
    beta_true = beta_true,
    true_idx = true_idx,
    DF = DF,
    rho = if (is.null(rho)) NULL else rho,
    args = list(
      n = n, p = p, s0 = s0, sigma = sigma,
      magnitudes_range = magnitudes_range,
      rho = rho, seed = seed
    )
  )
}
```

```{r}
sim <- sim_subset_data(n = 100, p = 50, s0 = 25, sigma = 1.5, rho = NULL, seed = 123)
y <- sim$y
X <- sim$X
ga <- gareg_subset(
  y = y, X = X, gaMethod = "GA", monitor = FALSE,
  gacontrol = list(
    popSize = 120,
    maxiter = 20000,
    run = 4000,
    pmutation = 0.02
  )
)
summary(ga)

res <- FDRCalc(truelabel = sim$true_idx, predlabel = ga@bestsol, N = 50)
# FALSE Discover Rate
res$fdr
# TRUE Positive Rate
res$tpr
```

## Changepoint detection

The multiple changepoint detection can be conducted through `changepointGA` 
package (Li and Lu, 2024). The BIC penalized function is provided below for IID data. The related 
math details can be found in (Li et al., 2026).


```{r}
BIC.cpt <- function(chromosome, Xt) {
  m <- chromosome[1]
  tau <- chromosome[2:(2 + m - 1)]
  N <- length(Xt)

  if (m == 0) {
    mu.hat <- mean(Xt)
    sigma.hatsq <- sum((Xt - mu.hat)^2) / N
    BIC.obj <- 0.5 * N * log(sigma.hatsq) + 2 * log(N)
  } else {
    tau.ext <- c(1, tau, N + 1)
    seg.len <- diff(tau.ext)
    ff <- rep(0:m, times = seg.len)
    Xseg <- split(Xt, ff)
    mu.seg <- unlist(lapply(Xseg, mean), use.names = F)
    mu.hat <- rep(mu.seg, seg.len)
    sigma.hatsq <- sum((Xt - mu.hat)^2) / N
    BIC.obj <- 0.5 * N * log(sigma.hatsq) + (2 * m + 2) * log(N)
  }
  return(BIC.obj)
}

# IID data
set.seed(1234)
n <- 200
et <- rnorm(n)
Xt <- et + rep(c(0, 2, 0, 2), each = n / 4)

library(changepointGA)
GA.res <- cptga(
  ObjFunc = BIC.cpt, N = n, popSize = 500,
  pcrossover = 0.95, pmutation = 0.3, pchangepoint = 10 / n,
  Xt = Xt
)
summary(GA.res)
```

## Optimal spline knot selection

### Example data set

This section will illustrate optimal spline knot placement on the classic motocycle impact dataset from Package \code{MASS} (Venables & Ripley, 2002).\code{MASS::mcycle} contains 133 observations from a simulated motorcycle crash test, recording head acceleration (in g) of a helmeted test subject over time (milliseconds). The series is non-linear with sharp transients and heteroskedastic noise, which makes it a canonical benchmark for smoothing and spline-based regression. We use function \code{gareg_knots} choose interior spline knots subject to a minimum separation in indices to avoid clustering. Here the `truncated-power piecewise polynomials` with polynomial degree of 3 is the used default.


```{r}
library(MASS)
library(splines)

data(mcycle)
head(mcycle)
```

### Different GA model set-up

- Varying knots with single population GA;
```{r}
g1 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  minDist = 5,
  gaMethod = "cptga",
  cptgactrl = cptgaControl(popSize = 200, pcrossover = 0.9, pmutation = 0.3),
  ic_method = "BIC"
)
summary(g1)

# knots location
g1@bestsol
```

- Varying knots with island model GA;
```{r}
g2 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  minDist = 5,
  gaMethod = "cptgaisl",
  cptgactrl = cptgaControl(
    numIslands = 5, popSize = 200, maxMig = 250,
    pcrossover = 0.9, pmutation = 0.3
  ),
  ic_method = "BIC"
)
summary(g2)
```

- Fixed knots with single population GA;
```{r}
g3 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  fixedknots = 3,
  minDist = 5,
  gaMethod = "cptga",
  cptgactrl = cptgaControl(popSize = 200, pcrossover = 0.9, pmutation = 0.3),
  ic_method = "BIC"
)
summary(g3)
```

- Fixed knots with island model GA;
```{r}
g4 <- gareg_knots(
  y = mcycle$accel, x = mcycle$times,
  fixedknots = 4,
  minDist = 5,
  gaMethod = "cptgaisl",
  cptgactrl = cptgaControl(
    numIslands = 5, popSize = 200, maxMig = 250,
    pcrossover = 0.9, pmutation = 0.3
  ),
  ic_method = "BIC"
)
summary(g4)
```

```{r, fig.width=7, fig.height=5, out.width="90%", fig.align='center'}
y <- mcycle$accel
x <- mcycle$times
x_unique <- unique(x)

tBIC.vary.ga <- g1@bestsol
tBIC.vary.gaisl <- g2@bestsol
tBIC.fix.3.ga <- g3@bestsol
tBIC.fix.4.gaisl <- g4@bestsol

bsfit.vary.ga <- lm(y ~ bs(x, knots = x_unique[g1@bestsol], Boundary.knots = range(x)))
bsfit.vary.gaisl <- lm(y ~ bs(x, knots = x_unique[g2@bestsol], Boundary.knots = range(x)))
bsfit.fix.3.ga <- lm(y ~ bs(x, knots = x_unique[g3@bestsol], Boundary.knots = range(x)))
bsfit.fix.4.gaisl <- lm(y ~ bs(x, knots = x_unique[g4@bestsol], Boundary.knots = range(x)))

plot(x, y, xlab = "Time (ms)", ylab = "Acceleration (g)")
ht <- seq(min(x), max(x), length.out = 200)
lines(ht, predict(bsfit.vary.ga, data.frame(x = ht)), col = "blue", lty = 5, lwd = 2)
lines(ht, predict(bsfit.vary.gaisl, data.frame(x = ht)), col = "orange", lty = 4, lwd = 2)
lines(ht, predict(bsfit.fix.3.ga, data.frame(x = ht)), col = "purple", lty = 3, lwd = 2)
lines(ht, predict(bsfit.fix.4.gaisl, data.frame(x = ht)), col = "#D55E00", lty = 2, lwd = 2)
legend("bottomright",
  legend = c(
    "Varying knots GA",
    "Varying knots island model GA",
    "Fixed 3 knots GA",
    "Fixed 4 knots island model GA"
  ),
  lty = 5:2, lwd = 2,
  col = c("blue", "orange", "purple", "#D55E00"), bty = "n"
)
```

### Spline options: piecewise polynomials, natural cubic, and B-splines

This section illustrates how to build spline design matrices via \code{splineX()} for three common options:

- \code{type="ppolys"}: degree-$d$ truncated power piecewise polynomials;
- \code{type="ns"}: degree-3 natural cubic spline (degree will be ignored);
- \code{type="bs"}: degree-$d$ B-spline basis;

We’ll use the motorcycle acceleration data \code{MASS::mcycle}, create interior knots at quantiles of \code{times}, and compare how different spline types/degrees behave. Here, we only illustrate through `Varying number and locations of knots` set-up (Let GA choose both **how many** knots and **where** they go). 

```{r vary-ppolys}
g_pp3 <- gareg_knots(
  y = y, x = x,
  minDist = 5,
  gaMethod = "cptga",
  ObjFunc = NULL, # use default varyknotsIC
  type = "ppolys",
  degree = 3, # degree-3 piecewise polynomial
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_pp3)
```

```{r vary-ns}
g_ns <- gareg_knots(
  y = y, x = x,
  minDist = 5,
  gaMethod = "cptga",
  type = "ns", # natural cubic (degree ignored)
  degree = 3, # ignored for "ns"
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_ns)
```


```{r vary-bs}
g_bs1 <- gareg_knots(
  y = y, x = x,
  minDist = 5,
  gaMethod = "cptga",
  type = "bs",
  degree = 1, # linear B-splines
  intercept = TRUE,
  ic_method = "BIC"
)
summary(g_bs1)
```


## Reproducibility

```{r}
sessionInfo()
```

## References

Li, M., & Lu, Q. (2025). changepointGA: An R package for Fast Changepoint Detection via Genetic Algorithm. arXiv preprint arXiv:2410.15571.

Mo Li, QiQi Lu, Robert Lund, & Xueheng Shi. (2026). Genetic Algorithms in Regression. In preparation.

Venables, W. N. & Ripley, B. D. (2002) Modern Applied Statistics with S. Fourth Edition. Springer, New York. ISBN 0-387-95457-0.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{cptgaControl}
\alias{cptgaControl}
\title{Build Control List for \code{cptga}/\code{cptgaisl}}
\usage{
cptgaControl(
  ...,
  .list = NULL,
  .persist = FALSE,
  .env = asNamespace("GAReg"),
  .validate = TRUE,
  engine = NULL
)
}
\arguments{
\item{...}{Named overrides for control fields (e.g., \code{popSize},
\code{pcrossover}, \code{minDist}, \code{numIslands}).}

\item{.list}{Optional named list of overrides (merged with \code{...}).}

\item{.persist}{Logical; if \code{TRUE}, persist updated defaults back into
the target environment (not usually recommended in user code).}

\item{.env}{Environment where defaults live (defaults to \code{parent.frame()}).}

\item{.validate}{Logical; validate values/ranges (default \code{TRUE}).}

\item{engine}{Character; one of \code{"cptga"} or \code{"cptgaisl"} to select
the default set and validation rules.}
}
\value{
A list of class \code{"cptgaControl"}.
}
\description{
Convenience constructor for GA control parameters used by
\code{changepointGA::cptga} and \code{changepointGA::cptgaisl}. It merges
named overrides into engine-specific defaults
(\link{.cptga.default} or \link{.cptgaisl.default}), with light validation.
}
\details{
Unknown names are rejected. When both \code{...} and \code{.list} are present,
they are combined, with later entries overwriting earlier ones.
}
\seealso{
\link{gareg_knots}, \link{.cptga.default}, \link{.cptgaisl.default}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-nhtr2005.R
\docType{data}
\name{nhtr2005}
\alias{nhtr2005}
\title{Northern Hemisphere temperature reconstruction}
\format{
A data frame with 2 variables:
\describe{
  \item{Year}{Year AD}
  \item{Temp}{Temperature anomaly relative to the 1961--1990 mean}
}
}
\source{
Moberg A, Sonechkin DM, Holmgren K, Datsenko NM, Karlén W.
  2,000-Year Northern Hemisphere Temperature Reconstruction. IGBP
  PAGES/World Data Center for Paleoclimatology Data Contribution Series
  #2005-019. NOAA/NGDC Paleoclimatology Program, Boulder, Colorado, USA.
}
\usage{
nhtr2005
}
\description{
The 2,000-Year Northern Hemisphere Temperature Reconstruction dataset
(Moberg et al., 2005) provides annual temperature anomalies for the
Northern Hemisphere from AD 1 to 1979, relative to the 1961–1990 mean.
The reconstruction combines high-resolution proxy data (e.g., tree rings)
with low-resolution proxies (e.g., sediments) using a wavelet-based method
to capture variability across multiple time scales.
}
\references{
Moberg A, Sonechkin DM, Holmgren K, Datsenko NM, Karlén W.
  (2005). Highly variable Northern Hemisphere temperatures reconstructed
  from low- and high-resolution proxy data. *Nature*, 433(7026), 613--617.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_subset.R
\name{FDRCalc}
\alias{FDRCalc}
\title{False Discovery Rate (FDR) and True Positive Rate (TPR) from index labels}
\usage{
FDRCalc(truelabel, predlabel, N)
}
\arguments{
\item{truelabel}{Integer vector of ground-truth positive indices (values in \code{1..N}).}

\item{predlabel}{Integer vector of predicted positive indices (values in \code{1..N}).}

\item{N}{Integer scalar; size of the full index universe (total number of candidates).}
}
\value{
A named list with components:
\itemize{
  \item \code{fdr} False Discovery Rate, \eqn{fp/(fp+tp)} (NaN if \code{fp+tp == 0}).
  \item \code{tpr} True Positive Rate (recall), \eqn{tp/(tp+fn)} (NaN if \code{tp+fn == 0}).
  \item \code{fp}, \code{fn}, \code{tp}, \code{tn} Confusion-matrix counts.
}
}
\description{
Computes the False Discovery Rate (FDR) and True Positive Rate (TPR, a.k.a. recall)
by comparing a set of \emph{true} labels to a set of \emph{predicted} labels.
Labels are treated as positive integer indices in \eqn{\{1, \dots, N\}}. Duplicates
are ignored (unique indices are used).
}
\details{
Let \code{truelabel} and \code{predlabel} be sets of indices. The function
derives the confusion-matrix counts:
\itemize{
  \item \code{tp} = \eqn{|truelabel \cap predlabel|}
  \item \code{fp} = \eqn{|predlabel \setminus truelabel|}
  \item \code{fn} = \eqn{|truelabel \setminus predlabel|}
  \item \code{tn} = \eqn{N - tp - fp - fn}
}
and returns
\deqn{FDR = fp / (fp + tp), \qquad TPR = tp / (tp + fn).}

Inputs are coerced to integer and uniqued. A warning is emitted if any label
is < 1, and an error is thrown if any label exceeds \code{N}. If \code{tn < 0},
a warning is issued indicating that \code{N} may not reflect the full universe.
}
\section{Edge cases}{

If \code{predlabel} is empty, \code{fdr} is \code{NaN} and \code{tpr} is 0 (unless
\code{truelabel} is also empty, in which case both \code{fdr} and \code{tpr} are \code{NaN}).
If \code{truelabel} is empty and \code{predlabel} non-empty, \code{tpr} is \code{NaN}
and \code{fdr} is 1.
}

\examples{
# Simple example
N <- 10
true <- c(2, 4, 7)
pred <- c(4, 5, 7, 7) # duplicates are ignored
FDRCalc(true, pred, N)

# Empty predictions
FDRCalc(true, integer(0), N)

# All correct predictions
FDRCalc(true, true, N)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg-class.R
\docType{class}
\name{gareg-class}
\alias{gareg-class}
\title{S4 Class Definition for `gareg`}
\description{
S4 Class for Genetic Algorithm-Based Regression

S4 container for GA-based regression/changepoint tasks. Holds the GA
backend fit and a normalized summary of the best solution.
}
\section{Slots}{

\describe{
\item{\code{call}}{The matched call that created the object.}

\item{\code{N}}{The effective size of the x grid used for knot search
(i.e., `length(x_unique)`), typically the number of unique `x`.}

\item{\code{call}}{language. The original call.}

\item{\code{method}}{character. One of "varyknots", "fixknots", "subset".}

\item{\code{N}}{numeric. Length of `x_unique` used by the GA (also `sentinel-1`).}

\item{\code{objFunc}}{functionOrNULL. Objective function used.}

\item{\code{gaMethod}}{character. GA engine name ("cptga","cptgaisl","ga","gaisl").}

\item{\code{gaFit}}{Backend GA fit object (union of classes from GA and changepointGA).}

\item{\code{ctrl}}{listOrNULL. Control list used to run the GA (if stored by caller).}

\item{\code{fixedknots}}{numericOrNULL. Fixed number of interior knots (`m`) for fixed-knots mode, or NULL.}

\item{\code{minDist}}{numeric. Minimum distance between adjacent changepoints.}

\item{\code{polydegree}}{numericOrNULL. Spline degree for default objectives.}

\item{\code{type}}{character. One of `c("ppolys", "ns", "bs")` indicating piecewise
polynomials, natural cubic, or B-spline.}

\item{\code{intercept}}{logical. Whether the spline basis included an intercept column.}

\item{\code{subsetSpec}}{listOrNULL. Constraints for subset selection (unused for knots).}

\item{\code{featureNames}}{character. Candidate feature names (subset tasks).}

\item{\code{bestFitness}}{numeric. Best fitness value found.}

\item{\code{bestChrom}}{numeric. Raw best chromosome returned by the backend
(may include a sentinel equal to `N+1` and optional padding).}

\item{\code{bestnumbsol}}{numeric. Count of selected elements (e.g., `m` for knots).}

\item{\code{bestsol}}{numericOrChara. For knots: the `m` interior indices (pre-sentinel);
for subset: mask/indices/names.}
}}

\seealso{
[gareg_knots], [cptgaControl]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg-methods.R
\name{gareg-methods}
\alias{gareg-methods}
\alias{show,gareg-method}
\alias{summary,gareg-method}
\title{Show and summary methods for \code{gareg}}
\usage{
\S4method{show}{gareg}(object)

\S4method{summary}{gareg}(object, ...)
}
\arguments{
\item{object}{A \code{"gareg"} object.}

\item{...}{Currently unused.}
}
\value{
\code{show}: invisible NULL. \code{summary}: invisibly returns \code{object}.
}
\description{
\itemize{
  \item \code{show(object)}: Compact header with call, engine, and N.
  \item \code{summary(object, ...)}: GA settings (when available) and best solution.
}
}
\details{
Methods for displaying and summarizing `gareg` objects
}
\seealso{
\link{gareg-class}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{splineX}
\alias{splineX}
\title{Build spline design matrices (piecewise polynomials, natural cubic, B-spline)}
\usage{
splineX(
  x,
  knots,
  degree = NULL,
  type = c("ppolys", "ns", "bs"),
  intercept = TRUE
)
}
\arguments{
\item{x}{Numeric vector of predictor values.}

\item{knots}{Numeric vector of interior knots.}

\item{degree}{Integer polynomial degree for \code{"ppolys"} and \code{"bs"}.
Ignored for \code{"ns"} (always cubic). Must be provided for
\code{"ppolys"} and \code{"bs"}.}

\item{type}{One of \code{c("ppolys", "ns", "bs")}.}

\item{intercept}{Logical; include intercept column where applicable. Default: `TRUE`.}
}
\value{
A numeric design matrix. Attributes are attached:
  \itemize{
    \item \code{"knots"} — the interior knots used
    \item \code{"boundary"} — \code{range(x)}
    \item \code{"degree"} — effective degree (i.e., 3 for \code{"ns"})
    \item \code{"type"} — the requested spline type
  }
}
\description{
Unified wrapper to generate spline covariates for three common cases:
\itemize{
  \item \code{type = "ppolys"}: Degree-\eqn{d} regression spline via
        truncated-power **piecewise polynomials** (uses internal \code{tp_basis()}).
  \item \code{type = "ns"}: Degree-3 **natural cubic spline**; enforces
        \eqn{f''(a)=f''(b)=0} at the boundary.
  \item \code{type = "bs"}: Degree-\eqn{d} **B-spline** basis (unconstrained).
}
}
\details{
Knots are sorted, no-duplicated, and any knots outside \code{range(x)} are
dropped with a warning. For \code{type = "ns"}, \code{degree} is ignored
(natural splines are cubic).
}
\examples{
set.seed(1)
x <- sort(rnorm(100))
k <- quantile(x, probs = c(.25, .5, .75))

# 1) Piecewise polynomials (degree 3)
X_pp <- splineX(x, knots = k, degree = 3, type = "ppolys", intercept = TRUE)
dim(X_pp) # n x ((3+1) + 3) = n x 7

# 2) Natural cubic spline (cubic, degree ignored)
X_ns <- splineX(x, knots = k, type = "ns", intercept = TRUE)

# 3) B-spline basis (degree 3)
X_bs <- splineX(x, knots = k, degree = 3, type = "bs", intercept = TRUE)

# Fit without a duplicated intercept:
# fit <- lm(y ~ 0 + X_pp)

}
\seealso{
\code{\link[splines]{bs}}, \code{\link[splines]{ns}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{gareg_knots}
\alias{gareg_knots}
\title{Genetic-Algorithm–based Optimal Knot Selection}
\usage{
gareg_knots(
  y,
  x,
  ObjFunc = NULL,
  fixedknots = NULL,
  minDist = 3L,
  degree = 3L,
  type = c("ppolys", "ns", "bs"),
  intercept = TRUE,
  gaMethod = "cptga",
  cptgactrl = NULL,
  monitoring = FALSE,
  seed = NULL,
  ...
)
}
\arguments{
\item{y}{Numeric vector of responses (length \code{N}).}

\item{x}{Optional index/time vector aligned with \code{y}. If missing, it
defaults to \code{seq_along(y)}. Used to derive \code{x_unique} (candidate
knot positions) and passed to the objective function; the GA backend itself
does not use \code{x} directly.}

\item{ObjFunc}{Objective function or its name. If \code{NULL}, a default
is chosen:
\itemize{
  \item \code{fixknotsIC} when \code{fixedknots} is supplied;
  \item \code{varyknotsIC} otherwise.
}
A custom function must accept the chromosome and needed data via
named arguments (see the defaults for a template function).}

\item{fixedknots}{\code{NULL} (varying-knots search) or an integer giving the
number of interior knots for a fixed-\eqn{m} search. If non-\code{NULL}, the
method is \code{"fixknots"} and specialized operators are injected unless
overridden in \code{cptgactrl}.}

\item{minDist}{Integer minimum distance between adjacent changepoints.
If omitted (\code{missing()} or \code{NULL}), the value in \code{cptgactrl}
is used. If supplied here, it overrides the control value.}

\item{degree}{Integer polynomial degree for \code{"ppolys"} and \code{"bs"}.
Ignored for \code{"ns"} (always cubic). Must be provided for \code{"ppolys"}
and \code{"bs"}.}

\item{type}{One of \code{c("ppolys", "ns", "bs")}:
piecewise polynomials, natural cubic, or B-spline. See \code{\link{splineX}}.
The first option of `ppolys` is taken by default.}

\item{intercept}{Logical; include intercept column where applicable.
Default: \code{TRUE}.}

\item{gaMethod}{GA backend to call: function or name. Supports
\code{"cptga"} (single population) and \code{"cptgaisl"} (islands).}

\item{cptgactrl}{Control list built with \code{\link{cptgaControl}()}
(or a named list of overrides). When \code{gaMethod = "cptgaisl"},
island-specific knobs like \code{numIslands} and \code{maxMig} are
recognized. Other genetic algorithm parameters can be found in \link[changepointGA:cptga]{cptga} and
      \link[changepointGA:cptgaisl]{cptgaisl}.}

\item{monitoring}{Logical; print short progress messages (also forwarded
into the backend control).}

\item{seed}{Optional RNG seed; also stored into the backend control.}

\item{...}{Additional arguments passed to the GA backend. If the backend
does not accept \code{...}, unknown arguments are silently dropped
(the call is filtered against the backend formals).}
}
\value{
An object of class \code{"gareg"} with key slots:
\itemize{
  \item \code{call}, \code{method} (\code{"varyknots"} or \code{"fixknots"}), \code{N}.
  \item \code{objFunc}, \code{gaMethod}, \code{gaFit} (class \code{"cptga"} or \code{"cptgaisl"}), \code{ctrl}.
  \item \code{fixedknots}, \code{minDist}, \code{polydegree}, \code{type}, \code{intercept}.
  \item \code{bestFitness}, \code{bestChrom}, \code{bestnumbsol}, \code{bestsol}.
}
Use \code{summary(g)} to print GA settings and the best solution (extracted
from \code{g@gaFit}); \code{show(g)} prints a compact header.
}
\description{
Runs a GA-based search for changepoints/knots and returns a compact
\code{"gareg"} S4 result that stores the backend GA fit
(\code{"cptga"} or \code{"cptgaisl"}) plus the essential run settings.
}
\details{
\strong{Engine selection and controls.}
The function detects the engine from \code{gaMethod} and constructs a
matching control via \code{\link{cptgaControl}()}:
\itemize{
  \item \code{"cptga"} uses \code{.cptga.default}.
  \item \code{"cptgaisl"} uses \code{.cptgaisl.default} (supports
        \code{numIslands}, \code{maxMig}, etc.).
  \item see other details in \link[changepointGA:cptga]{cptga} and
        \link[changepointGA:cptgaisl]{cptgaisl}.
}
Top-level \code{monitoring}, \code{seed}, and \code{minDist} given to
\code{gareg_knots()} take precedence over the control list.

\strong{Fix-knots operators.}
When \code{fixedknots} is provided and the control does not already
override them, the following operators are injected:
\code{Popinitial_fixknots}, \code{crossover_fixknots}, \code{mutation_fixknots}.

\strong{Spline basis options.}
To build spline design matrices (via \code{\link{splineX}}):
\itemize{
  \item \code{type = "ppolys"}: Degree-\eqn{d} regression spline via truncated-power piecewise polynomials.
  \item \code{type = "ns"}: Degree-3 natural cubic spline with zero second-derivative at boundaries.
  \item \code{type = "bs"}: Degree-\eqn{d} B-spline basis (unconstrained).
}
}
\section{Argument precedence}{

Values are combined as \emph{control < core < \code{...}}. That is,
\code{cptgactrl} provides defaults, then core arguments from
\code{gareg_knots()} override those, and finally any matching names in
\code{...} override both.
}

\examples{
\donttest{
set.seed(1)
N <- 120
y <- c(rnorm(40, 0), rnorm(40, 3), rnorm(40, 0))
x <- seq_len(N)

# 1) Varying-knots with single-pop GA
g1 <- gareg_knots(
  y, x,
  minDist = 5,
  gaMethod = "cptga",
  cptgactrl = cptgaControl(popSize = 150, pcrossover = 0.9, maxgen = 500)
)
summary(g1)

# 2) Fixed knots (operators auto-injected unless overridden)
g2 <- gareg_knots(
  y, x,
  fixedknots = 5,
  minDist = 5
)
summary(g2)

# 3) Island GA with island-specific controls
g3 <- gareg_knots(
  y, x,
  gaMethod = "cptgaisl",
  minDist = 6,
  cptgactrl = cptgaControl(
    engine = "cptgaisl",
    numIslands = 8, maxMig = 250,
    popSize = 120, pcrossover = 0.9
  )
)
summary(g3)
}

}
\seealso{
\code{\link{cptgaControl}}, \code{changepointGA::cptga},
  \code{changepointGA::cptgaisl}, \code{\link{fixknotsIC}}, \code{\link{varyknotsIC}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\docType{data}
\name{.cptga.default}
\alias{.cptga.default}
\title{Default Controls for \code{cptga}}
\format{
A named list with fields like \code{popSize}, \code{pcrossover},
\code{pmutation}, \code{pchangepoint}, \code{minDist}, \code{maxgen},
\code{option}, \code{selection}, \code{crossover}, \code{mutation}, etc.
}
\usage{
.cptga.default
}
\description{
Engine defaults used by \link{cptgaControl} when \code{engine = "cptga"}.
Not exported; shown here for reference.
}
\seealso{
\link{cptgaControl}, \link{.cptgaisl.default}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{fixknotsIC}
\alias{fixknotsIC}
\title{Information criterion for a fixed–knot spline regression}
\usage{
fixknotsIC(
  knot_bin,
  plen = 0,
  y,
  x,
  x_unique,
  x_base = NULL,
  fixedknots,
  degree = 3L,
  type = c("ppolys", "ns", "bs"),
  intercept = TRUE,
  ic_method = "BIC"
)
}
\arguments{
\item{knot_bin}{Integer vector (chromosome). Gene 1 stores m, the number of
interior knots. Genes 2:(1+m) are indices into \code{x_unique} selecting the m
interior knots, followed by a sentinel equal to \code{length(x_unique)+1}.
Only genes strictly before the first occurrence of \code{length(x_unique)+1}
are treated as interior indices; genes after the sentinel are ignored. Interior
indices must be in \code{2:(length(x_unique)-1)}, finite, and non-duplicated.}

\item{plen}{Unused placeholder kept for API compatibility with other
objective functions. Ignored.}

\item{y}{Numeric response vector of length \eqn{n}.}

\item{x}{Numeric predictor (same length as \code{y}) on which the spline
basis is built.}

\item{x_unique}{Optional numeric vector of unique candidate knot locations.
If \code{NULL} or missing, it defaults to \code{sort(unique(x))}. Must
contain at least \eqn{m + 2} values (interior + two boundaries).}

\item{x_base}{Optional matrix (or vector) of additional covariates to include
linearly alongside the spline basis. If supplied, it is coerced to a matrix
and column-bound to the design.}

\item{fixedknots}{Integer \eqn{m}: the number of interior knots to use.
Internally this determines how many indices are read from \code{knot_bin}.}

\item{degree}{Integer polynomial degree for \code{type="ppolys"} and
\code{type="bs"} (default \code{3L}). Ignored for \code{type="ns"} (cubic).}

\item{type}{One of \code{c("ppolys","ns","bs")}; forwarded to [splineX()].}

\item{intercept}{Logical; forwarded to [splineX()]. For \eqn{m>0}, the spline
block is \code{splineX(..., intercept=intercept)} and no explicit 1-column is
added here. If you add your own intercept in \code{X}, call
\code{splineX(..., intercept=FALSE)}.}

\item{ic_method}{Character; which information criterion to return:
\code{"BIC"}, \code{"AIC"}, or \code{"AICc"}.}
}
\value{
A single numeric value: the requested information criterion. Lower is
  better. Returns \code{Inf} for invalid chromosomes/inputs.
}
\description{
Computes an information criterion (BIC, AIC, or AICc) for a regression of
\code{y} on a spline basis of \code{x} when the number of interior knots is
fixed. This is designed to be used as a fitness/objective function inside a
GA search where the chromosome encodes the indices of the interior knots.
}
\details{
We decode the interior indices up to the sentinel \code{length(x_unique)+1},
validate them (finite, interior, non-duplicated), sort the resulting knot
locations internally, and build the design as
\code{X <- cbind(splineX(..., intercept=intercept), x_base)}.
Invalid chromosomes/inputs return \code{Inf}.
}
\examples{
library(MASS)
y <- mcycle$accel
x <- mcycle$times
x_unique <- sort(unique(x))
# chromosome encoding 5 interior knot indices with sentinel:
chrom <- c(5, 24, 30, 46, 49, 69, length(x_unique) + 1)
fixknotsIC(chrom,
  y = y, x = x, x_unique = x_unique,
  fixedknots = 5, ic_method = "BIC"
)
}
\seealso{
[varyknotsIC()], [splineX()], \code{\link[splines]{bs}}, \code{\link[splines]{ns}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_subset.R
\name{gareg_subset}
\alias{gareg_subset}
\title{Genetic-Algorithm Best Subset Selection (GA / GAISL)}
\usage{
gareg_subset(
  y,
  X,
  ObjFunc = NULL,
  gaMethod = "ga",
  gacontrol = NULL,
  monitoring = FALSE,
  seed = NULL,
  ...
)
}
\arguments{
\item{y}{Numeric response vector (length \code{n}).}

\item{X}{Numeric matrix of candidate predictors (\code{n} rows by \code{p} columns).}

\item{ObjFunc}{Objective function or its name. Defaults to \code{\link{subsetBIC}}.
The objective must accept as its first argument a binary chromosome
(0/1 mask of length \code{p}) and may accept additional arguments passed via \code{...}.
By convention, \code{subsetBIC} returns \emph{negative} BIC, so the GA maximizes fitness.}

\item{gaMethod}{GA backend to call: \code{"ga"} or \code{"gaisl"} (functions from package \pkg{GA}),
or a GA-compatible function with the same interface as \code{\link[GA]{ga}}.}

\item{gacontrol}{Optional named list of GA engine controls (e.g., \code{popSize}, \code{maxiter},
\code{run}, \code{pcrossover}, \code{pmutation}, \code{elitism}, \code{seed}, \code{parallel},
\code{keepBest}, \code{monitor}, ...). These are passed to the GA engine, not to the objective.}

\item{monitoring}{Logical; if \code{TRUE}, prints a short message and (if not supplied in
\code{gacontrol}) sets \code{monitor = GA::gaMonitor} for live progress.}

\item{seed}{Optional RNG seed (convenience alias for \code{gacontrol$seed}).}

\item{...}{Additional arguments forwarded to \code{ObjFunc} (not to the GA engine). For
\code{\link{subsetBIC}} these typically include \code{family}, \code{weights}, \code{offset},
and \code{control}.}
}
\value{
An object of S4 class \code{"gareg"} (with \code{method = "subset"}) containing:
\itemize{
  \item \code{call} – the matched call.
  \item \code{N} – number of observations.
  \item \code{objFunc} – the objective function used.
  \item \code{gaMethod} – \code{"ga"} or \code{"gaisl"}.
  \item \code{gaFit} – the GA fit object returned by \pkg{GA} (if your class allows it).
  \item \code{featureNames} – column names of \code{X} (or empty).
  \item \code{bestFitness} – best fitness value (\code{GA::ga}@fitnessValue).
  \item \code{bestChrom} – \code{c(m, idx)}: number of selected variables and their indices.
  \item \code{bestnumbsol} – \code{m}, number of selected variables.
  \item \code{bestsol} – vector of selected column indices in \code{X}.
}
}
\description{
Runs a GA-based search over variable subsets using a user-specified
objective (default: \code{\link{subsetBIC}}) and returns a compact
\code{"gareg"} S4 result with \code{method = "subset"}.
The engine can be \code{\link[GA]{ga}} (single population) or
\code{\link[GA]{gaisl}} (islands), selected via \code{gaMethod}.
}
\details{
The fitness passed to \pkg{GA} is \code{ObjFunc} itself. Because the engine expects
a function with signature \code{f(chrom, ...)}, your \code{ObjFunc} must interpret
\code{chrom} as a 0/1 mask over the columns of \code{X}. The function then computes a score
(e.g., negative BIC) using \code{y}, \code{X}, and any extra arguments supplied via \code{...}.

With the default \code{\link{subsetBIC}}, the returned value is \code{-BIC}, so we set
\code{max = TRUE} in the GA call to maximize fitness. If you switch to an objective that
returns a quantity to \emph{minimize}, either negate it in your objective or change
the engine setting to \code{max = FALSE}.

Engine controls belong in \code{gacontrol}; objective-specific options belong in \code{...}.
This separation prevents accidental name collisions between GA engine parameters and
objective arguments.
}
\examples{
\donttest{
if (requireNamespace("GA", quietly = TRUE)) {
  set.seed(1)
  n <- 100
  p <- 12
  X <- matrix(rnorm(n * p), n, p)
  y <- 1 + X[, 1] - 0.7 * X[, 4] + rnorm(n, sd = 0.5)

  # Default: subsetBIC (Gaussian – negative BIC), engine = GA::ga
  fit1 <- gareg_subset(y, X,
    gaMethod = "ga",
    gacontrol = list(popSize = 60, maxiter = 80, run = 40, parallel = FALSE)
  )
  summary(fit1)

  # Island model: GA::gaisl
  fit2 <- gareg_subset(y, X,
    gaMethod = "gaisl",
    gacontrol = list(popSize = 40, maxiter = 60, numIslands = 4, parallel = FALSE)
  )
  summary(fit2)

  # Logistic objective (subsetBIC handles GLM via ...):
  ybin <- rbinom(n, 1, plogis(0.3 + X[, 1] - 0.5 * X[, 2]))
  fit3 <- gareg_subset(ybin, X,
    gaMethod = "ga",
    family = stats::binomial(), # <- passed to subsetBIC via ...
    gacontrol = list(popSize = 60, maxiter = 80, parallel = FALSE)
  )
  summary(fit3)
}
}

}
\seealso{
\code{\link{subsetBIC}},
\code{\link[GA]{ga}},
\code{\link[GA]{gaisl}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{crossover_fixknots}
\alias{crossover_fixknots}
\title{Crossover Operator (Fixed-\eqn{m}) with Feasibility-First Restarts}
\usage{
crossover_fixknots(mom, dad, prange = NULL, minDist, lmax, N)
}
\arguments{
\item{mom, dad}{Integer vectors encoding parent chromosomes:
first entry \eqn{m} (number of changepoints), followed by \eqn{m} ordered
knot locations.}

\item{prange}{Unused placeholder (kept for compatibility with other GA
operators). Default \code{NULL}.}

\item{minDist}{Integer; minimum spacing between adjacent knots in the child.}

\item{lmax}{Integer; chromosome length (number of rows in the population
matrix).}

\item{N}{Integer; series length. Used to place the sentinel \code{N+1} at
position \code{m+2}.}
}
\value{
An integer vector of length \code{lmax} encoding the child chromosome:
\code{c(m, child_knots, N+1, 0, 0, ...)}.
}
\description{
Produces a child chromosome from two fixed-\eqn{m} parents (same number of
knots) by alternately sampling candidate knot locations from the parents and
enforcing the spacing constraint \code{diff(child) > minDist}. If a conflict
is encountered, the routine restarts the construction up to a small cap.
}
\details{
Let \code{mom} and \code{dad} be chromosomes of the form
\code{c(m, tau_1, ..., tau_m, ...)}. This operator:
\enumerate{
  \item Initializes an empty child of size \eqn{m}.
  \item Picks the first knot at random from \code{mom} or \code{dad}.
  \item For each subsequent position \eqn{i=2,\dots,m}, considers the
        pair \code{(mom[i], dad[i])} and chooses the first value that
        maintains the spacing constraint relative to the previously chosen
        knot (\code{> minDist}); if both work, one is chosen at random.
  \item If no feasible choice exists at some step, the construction restarts
        from the first position (up to a small cap governed internally by
        \code{up_tol}).
}
The result is written back as a full-length chromosome with the sentinel
\code{N+1} in position \code{m+2}, and zeros elsewhere.
}
\examples{
\donttest{
N <- 120
lmax <- 30
minDist <- 5
m <- 3
mom <- c(m, c(20, 50, 90), rep(0, lmax - 1 - m))
mom[m + 2] <- N + 1
dad <- c(m, c(18, 55, 85), rep(0, lmax - 1 - m))
dad[m + 2] <- N + 1
child <- crossover_fixknots(mom, dad, minDist = minDist, lmax = lmax, N = N)
child
}

}
\seealso{
\link{crossover_fixknots}, \link{mutation_fixknots},
\link{selectTau_uniform_exact}, \link{Popinitial_fixknots},
\link{gareg_knots}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_subset.R
\name{subsetBIC}
\alias{subsetBIC}
\title{Unified BIC-style Objective for Subset Selection (GLM & Gaussian)}
\usage{
subsetBIC(
  subset_bin,
  y,
  X,
  family = stats::gaussian(),
  weights = NULL,
  offset = NULL,
  control = stats::glm.control()
)
}
\arguments{
\item{subset_bin}{Integer/numeric 0–1 vector (length \code{ncol(X)}); 1 means
the corresponding column of \code{X} is included in the model.}

\item{y}{Numeric response vector of length \code{n}.}

\item{X}{Numeric matrix of candidate predictors; columns correspond to variables.}

\item{family}{A GLM family object (default \code{stats::gaussian()}).}

\item{weights}{Optional prior weights (passed to \code{glm.fit}).}

\item{offset}{Optional offset (passed to \code{glm.fit}).}

\item{control}{GLM fit controls; default \code{stats::glm.control()}.}
}
\value{
A single numeric value: \strong{-BIC}. Larger is better for GA maximizers.
  Returns \code{Inf} for rank-deficient designs.
}
\description{
Computes a BIC-like criterion for a chromosome that encodes a variable
subset. The same expression
\deqn{\mathrm{BIC} = n \log(\mathrm{rss\_like}/n) + k \log n}
is used for all families, where:
\itemize{
  \item For Gaussian with identity link, \code{rss_like} is the residual sum of squares (RSS),
        computed via a fast \code{.lm.fit}.
  \item For other GLM families, \code{rss_like} is the residual \emph{deviance}
        from \code{glm.fit}.
}
The effective parameter count \eqn{k} includes the intercept.
}
\details{
The chromosome \code{subset_bin} is a \emph{binary} vector (0/1 by column),
indicating which predictors from \code{X} are included. The design matrix
always includes an intercept. Rank-deficient selections return \code{Inf}
(which the GA maximizer treats as a very poor score). The value returned is
\strong{-BIC} so that GA engines can \emph{maximize} it.
}
\seealso{
\code{\link[stats]{glm.fit}}, \code{\link[stats]{glm.control}},
  \code{\link[stats]{.lm.fit}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{Popinitial_fixknots}
\alias{Popinitial_fixknots}
\title{Fixed-Knots Population Initializer}
\usage{
Popinitial_fixknots(
  popSize,
  prange = NULL,
  N,
  minDist,
  Pb,
  mmax,
  lmax,
  fixedknots
)
}
\arguments{
\item{popSize}{Integer; number of individuals (columns).}

\item{prange}{Optional hyperparameter range (unused here).}

\item{N}{Series length.}

\item{minDist}{Integer minimum spacing between adjacent changepoints.}

\item{Pb}{Unused placeholder (kept for compatibility).}

\item{mmax, lmax}{Integers; maximum number of knots and chromosome length.}

\item{fixedknots}{Integer; number of knots to place.}
}
\value{
Integer matrix of size \code{lmax x popSize}; each column is a
  chromosome \code{c(m, tau_1, ..., tau_m, N+1, ...)}.
}
\description{
Initializes a population matrix for the fixed-knots GA. Each column is a
feasible chromosome sampled by \link{selectTau_uniform_exact}.
}
\seealso{
\link{selectTau_uniform_exact}, \link{gareg_knots}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\docType{data}
\name{.cptgaisl.default}
\alias{.cptgaisl.default}
\title{Default Controls for \code{cptgaisl} (Island GA)}
\format{
A named list with fields like \code{popSize}, \code{numIslands},
\code{pcrossover}, \code{pmutation}, \code{maxMig}, \code{maxgen}, etc.
}
\usage{
.cptgaisl.default
}
\description{
Engine defaults used by \link{cptgaControl} when \code{engine = "cptgaisl"}.
Includes island-specific fields (e.g., \code{numIslands}, \code{maxMig}).
}
\seealso{
\link{cptgaControl}, \link{.cptga.default}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{varyknotsIC}
\alias{varyknotsIC}
\title{Information criterion for spline regression with a variable number of knots}
\usage{
varyknotsIC(
  knot_bin,
  plen = 0,
  y,
  x,
  x_unique,
  x_base = NULL,
  degree = 3L,
  type = c("ppolys", "ns", "bs"),
  intercept = TRUE,
  ic_method = "BIC"
)
}
\arguments{
\item{knot_bin}{Integer vector (chromosome). Gene 1 stores m, the number of
interior knots. Genes 2:(1+m) are indices into \code{x_unique} selecting the m
interior knots, followed by a sentinel equal to \code{length(x_unique)+1}.
Only genes strictly before the first occurrence of \code{length(x_unique)+1}
are treated as interior indices; genes after the sentinel are ignored. Interior
indices must be in \code{2:(length(x_unique)-1)}, finite, and non-duplicated.}

\item{plen}{Unused placeholder kept for API compatibility; ignored.}

\item{y}{Numeric response vector of length \eqn{n}.}

\item{x}{Numeric predictor (same length as \code{y}) on which the spline
basis is constructed.}

\item{x_unique}{Optional numeric vector of unique candidate knot locations.
If missing or \code{NULL}, defaults to \code{sort(unique(x))}. Must have
at least three values (two boundaries + one interior) to allow any knots.}

\item{x_base}{Optional matrix (or vector) of additional covariates to include
linearly alongside the spline basis; coerced to a matrix if supplied.}

\item{degree}{Integer polynomial degree for \code{type="ppolys"} and
\code{type="bs"} (default \code{3L}). Ignored for \code{type="ns"} (always cubic).}

\item{type}{One of \code{c("ppolys","ns","bs")}; forwarded to [splineX()].}

\item{intercept}{Logical; forwarded to [splineX()]. For \eqn{m>0}, the spline
block is \code{splineX(..., intercept=intercept)} and \emph{no explicit 1-column}
is added here; for \eqn{m=0}, an explicit intercept is added via
\code{cbind(1, x_base)}. Set \code{intercept=FALSE} if you plan to add your own 1-column.}

\item{ic_method}{Which information criterion to return: \code{"BIC"},
\code{"AIC"}, or \code{"AICc"}.}
}
\value{
A single numeric value: the requested information criterion (lower is
  better). Returns \code{Inf} for invalid chromosomes/inputs.
}
\description{
Evaluates an information criterion (BIC, AIC, or AICc) for a regression of
\code{y} on a spline basis of \code{x} where the number and locations of
interior knots are encoded in the chromosome. Designed for use as a GA
objective/fitness function. The spline basis is constructed via [splineX()].
}
\details{
If \eqn{m = 0}, the model is a pure-linear baseline using only an intercept
and \code{x_base}: \code{X <- cbind(1, x_base)} (no spline terms).
For \eqn{m > 0}, the spline block is built with [splineX()] using the selected
interior knots, with \code{X <- cbind(splineX(..., intercept=intercept), x_base)}.

The criteria are computed as:
\deqn{\mathrm{BIC} = n \log(\mathrm{SSRes}/n) + p \log n,}
\deqn{\mathrm{AIC} = n \log(\mathrm{SSRes}/n) + 2p,}
\deqn{\mathrm{AICc} = n \log(\mathrm{SSRes}/n) + 2p +
      \frac{2p(p+1)}{n-p-1},}
where \eqn{\mathrm{SSRes}} is the residual sum of squares and \eqn{p} is the
number of columns in the design matrix \code{X}.
}
\note{
This function allows \eqn{m=0} (no spline terms) so that the GA can
  compare against a pure-linear baseline (intercept + \code{x_base}).
  Spacing constraints (e.g., minimum distance between indices) should be
  enforced by the GA operators or an external penalty.
}
\examples{
## Example with 'mcycle' data (MASS)
# y <- mcycle$accel; x <- mcycle$times
# x_unique <- sort(unique(x))
# chrom <- c(5, 24, 30, 46, 49, 69, length(x_unique) + 1)
# varyknotsIC(chrom, y=y, x=x, x_unique=x_unique,
#             type="ppolys", degree=3, ic_method="BIC")
}
\seealso{
[fixknotsIC()], [splineX()], \code{\link[splines]{bs}}, \code{\link[splines]{ns}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{selectTau_uniform_exact}
\alias{selectTau_uniform_exact}
\title{Exact Uniform Sampler of Feasible Changepoints}
\usage{
selectTau_uniform_exact(N, m, minDist, lmax)
}
\arguments{
\item{N}{Integer series length.}

\item{m}{Integer number of changepoints to place.}

\item{minDist}{Integer minimum spacing between adjacent changepoints.}

\item{lmax}{Integer chromosome length.}
}
\value{
Integer vector length \code{lmax}:
  \code{c(m, tau_1, ..., tau_m, N+1, 0, 0, ...)}.
}
\description{
Samples \eqn{m} ordered changepoint indices uniformly from all feasible
configurations on \code{1:N} subject to a minimum spacing \code{minDist}.
Encodes the result as a chromosome for downstream GA operators.
}
\seealso{
\link{Popinitial_fixknots}, \link{mutation_fixknots}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gareg_knots.R
\name{mutation_fixknots}
\alias{mutation_fixknots}
\title{Mutation Operator (Fixed-Knots)}
\usage{
mutation_fixknots(child, p.range = NULL, minDist, Pb, lmax, mmax, N)
}
\arguments{
\item{child}{Current chromosome (its first entry defines \eqn{m}).}

\item{p.range, Pb}{Unused placeholders (kept for compatibility).}

\item{minDist}{Integer minimum spacing.}

\item{lmax, mmax}{Integers; chromosome length and maximum \eqn{m} (unused).}

\item{N}{Integer series length.}
}
\value{
New feasible chromosome with the same \eqn{m}.
}
\description{
Replaces a child with a fresh feasible sample having the same \eqn{m},
drawn by \link{selectTau_uniform_exact}.
}
\seealso{
\link{crossover_fixknots}
}
